from dataclasses import dataclass
from typing import List, Optional, Union

from valor.enums import AnnotationType, TaskType
from valor.schemas.filters import Filter


@dataclass
class EvaluationParameters:
    """
    Defines parameters for evaluation methods.

    Attributes
    ----------
    iou_thresholds_to_compute : Optional[List[float]]
        A list of floats describing which Intersection over Unions (IoUs) to use when calculating metrics (i.e., mAP).
    iou_thresholds_to_return: Optional[List[float]]
        A list of floats describing which Intersection over Union (IoUs) thresholds to calculate a metric for. Must be a subset of `iou_thresholds_to_compute`.
    label_map: Optional[List[List[List[str]]]]
        Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
    recall_score_threshold: float, default=0
        The confidence score threshold for use when determining whether to count a prediction as a true positive or not while calculating Average Recall.
    compute_pr_curves: bool
        A boolean which determines whether we calculate precision-recall curves or not.
    pr_curve_iou_threshold: float, optional
            The IOU threshold to use when calculating precision-recall curves for object detection tasks. Defaults to 0.5. Does nothing when compute_pr_curves is set to False or None.

    """

    task_type: TaskType

    # object detection
    convert_annotations_to_type: Optional[AnnotationType] = None
    iou_thresholds_to_compute: Optional[List[float]] = None
    iou_thresholds_to_return: Optional[List[float]] = None
    label_map: Optional[List[List[List[str]]]] = None
    recall_score_threshold: float = 0
    compute_pr_curves: bool = False
    pr_curve_iou_threshold: float = 0.5


@dataclass
class EvaluationRequest:
    """
    An evaluation request.

    Defines important attributes of the API's `EvaluationRequest`.

    Attributes
    ----------
    model_names : List[str]
        The list of models we want to evaluate by name.
    datum_filter : schemas.Filter
        The filter object used to define what the model(s) is evaluating against.
    parameters : EvaluationParameters
        Any parameters that are used to modify an evaluation method.
    """

    model_names: Union[str, List[str]]
    datum_filter: Filter
    parameters: EvaluationParameters

    def __post_init__(self):
        if isinstance(self.datum_filter, dict):
            self.datum_filter = Filter(**self.datum_filter)
        if isinstance(self.parameters, dict):
            self.parameters = EvaluationParameters(**self.parameters)



import copy

import pytest

from valor import Annotation, Datum, GroundTruth, Label, Prediction, enums
from valor.schemas import Float, Polygon


def test_datum():
    Datum(uid="123")
    Datum(uid="123", metadata={})
    Datum(uid="123", metadata={"name": 1})

    # test `__post_init__`
    with pytest.raises(TypeError):
        Datum(uid=123)  # type: ignore
    with pytest.raises(TypeError):
        Datum(uid="123", metadata=1)  # type: ignore
    with pytest.raises(TypeError):
        Datum(uid="123", metadata=[1])  # type: ignore


def test_annotation(bbox, polygon, raster, labels, metadata):
    # valid
    Annotation(
        task_type=enums.TaskType.OBJECT_DETECTION,
        bounding_box=bbox,
        labels=labels,
    )
    Annotation(
        task_type=enums.TaskType.OBJECT_DETECTION,
        polygon=polygon,
        labels=labels,
    )
    Annotation(
        task_type=enums.TaskType.OBJECT_DETECTION, raster=raster, labels=labels
    )
    Annotation(
        task_type=enums.TaskType.SEMANTIC_SEGMENTATION,
        raster=raster,
        labels=labels,
    )
    Annotation(
        task_type=enums.TaskType.OBJECT_DETECTION,
        labels=labels,
        bounding_box=bbox,
        polygon=polygon,
        raster=raster,
    )
    Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=labels)
    Annotation(
        task_type=enums.TaskType.CLASSIFICATION, labels=labels, metadata={}
    )
    Annotation(
        task_type=enums.TaskType.CLASSIFICATION,
        labels=labels,
        metadata=metadata,
    )
    Annotation(
        task_type=enums.TaskType.OBJECT_DETECTION,
        labels=labels,
        polygon=bbox,  # bbox is a constrained polygon so this is valid usage
    )

    # test `__post_init__`
    with pytest.raises(ValueError) as e:
        Annotation(task_type="something", labels=labels)  # type: ignore
    assert "TaskType" in str(e)
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.OBJECT_DETECTION,
            labels=labels,
            bounding_box=Polygon([[(0, 0), (1, 0), (1, 1), (0, 0)]]),  # type: ignore
        )
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.OBJECT_DETECTION,
            labels=labels,
            raster=bbox,
        )
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION,
            labels=labels,
            metadata=[1234],  # type: ignore
        )
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION,
            labels=labels,
            metadata={1: 1},  # type: ignore
        )
    with pytest.raises(ValueError):
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION,
            labels=labels,
            metadata={"test": None},  # type: ignore
        )


def test_groundtruth_annotation():
    l1 = Label(key="test", value="value")
    l2 = Label(key="test", value="other")
    l3 = Label(key="other", value="value")

    # valid
    Annotation(
        task_type=enums.TaskType.CLASSIFICATION,
        labels=[l1, l2, l3],
    )

    # test `__post_init__`
    with pytest.raises(ValueError) as e:
        Annotation(task_type="soemthing", labels=[l1])  # type: ignore
    assert "TaskType" in str(e)
    with pytest.raises(TypeError) as e:
        Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=l1)  # type: ignore
    assert "List[Label]" in str(e)
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION, labels=[l1, l2, "label"]  # type: ignore
        )


def test_prediction_annotation():
    l1 = Label(key="test", value="value")
    l2 = Label(key="test", value="other")
    l3 = Label(key="other", value="value")

    s1 = copy.deepcopy(l1)
    s1.score = Float.nullable(0.5)
    s2 = copy.deepcopy(l2)
    s2.score = Float.nullable(0.5)
    s3 = copy.deepcopy(l3)
    s3.score = Float.nullable(1.0)

    # valid
    Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=[s1, s2, s3])

    # test `__post_init__`
    with pytest.raises(ValueError) as e:
        Annotation(task_type="something", labels=[s1, s2, s3])  # type: ignore
    assert "TaskType" in str(e)
    with pytest.raises(TypeError) as e:
        Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=s1)  # type: ignore
    assert "List[Label]" in str(e)
    with pytest.raises(TypeError):
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION, labels=[s1, s2, "label"]  # type: ignore
        )


def test_groundtruth():
    label = Label(key="test", value="value")
    datum = Datum(uid="somefile")
    gts = [
        Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=[label]),
        Annotation(task_type=enums.TaskType.CLASSIFICATION, labels=[label]),
    ]

    # valid
    GroundTruth(
        datum=datum,
        annotations=gts,
    )

    # test `__post_init__`
    with pytest.raises(TypeError):
        GroundTruth(
            datum="datum",  # type: ignore
            annotations=gts,
        )
    with pytest.raises(TypeError) as e:
        GroundTruth(
            datum=datum,
            annotations=gts[0],  # type: ignore
        )
    assert "List[Annotation]" in str(e)
    with pytest.raises(TypeError):
        GroundTruth(
            datum=datum,
            annotations=[gts[0], gts[1], "annotation"],  # type: ignore
        )

    # test equalities
    with pytest.raises(TypeError):
        _ = (
            GroundTruth(
                datum=datum,
                annotations=gts,
            )
            == 1
        )

    assert GroundTruth(datum=datum, annotations=gts,) == GroundTruth(
        datum=datum,
        annotations=gts,
    )


def test_prediction():
    scored_label = Label(key="test", value="value", score=1.0)
    datum = Datum(uid="somefile")
    pds = [
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION,
            labels=[scored_label],
        ),
        Annotation(
            task_type=enums.TaskType.CLASSIFICATION,
            labels=[scored_label],
        ),
    ]

    # valid
    Prediction(datum=datum, annotations=pds)

    string = str(Prediction(datum=datum, annotations=pds))
    assert (
        string
        == "{'datum': {'uid': 'somefile', 'metadata': {}}, 'annotations': [{'task_type': 'classification', 'metadata': {}, 'labels': [{'key': 'test', 'value': 'value', 'score': 1.0}], 'bounding_box': None, 'polygon': None, 'raster': None, 'embedding': None}, {'task_type': 'classification', 'metadata': {}, 'labels': [{'key': 'test', 'value': 'value', 'score': 1.0}], 'bounding_box': None, 'polygon': None, 'raster': None, 'embedding': None}]}"
    )
    assert "dataset_name" not in string

    # test `__post_init__`
    with pytest.raises(TypeError):
        Prediction(datum="datum", annotations=pds)  # type: ignore
    with pytest.raises(TypeError) as e:
        Prediction(
            datum=datum,
            annotations=pds[0],  # type: ignore
        )
    assert "List[Annotation]" in str(e)

    with pytest.raises(TypeError):
        Prediction(
            datum=datum,
            annotations=[pds[0], pds[1], "annotation"],  # type: ignore
        )

    with pytest.raises(ValueError) as e:
        Prediction(
            datum=datum,
            annotations=[
                Annotation(
                    task_type=enums.TaskType.CLASSIFICATION,
                    labels=[
                        Label(key="test", value="value", score=0.8),
                        Label(key="test", value="other", score=0.1),
                    ],
                )
            ],
        )
    assert "for label key test got scores summing to 0.9" in str(e)

    # test equalities
    with pytest.raises(TypeError):
        _ = Prediction(datum=datum, annotations=pds) == 1

    assert Prediction(datum=datum, annotations=pds) == Prediction(
        datum=datum, annotations=pds
    )



import sys

if sys.version_info.minor >= 8:
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata  # type: ignore

try:
    __version__ = importlib_metadata.version("valor-client")
except importlib_metadata.PackageNotFoundError:
    __version__ = ""

from .client import ClientConnection, connect
from .coretypes import (
    Client,
    Dataset,
    Evaluation,
    GroundTruth,
    Model,
    Prediction,
)
from .schemas import Annotation, Datum, Filter, Label

__all__ = [
    "connect",
    "Client",
    "ClientConnection",
    "Evaluation",
    "Dataset",
    "Model",
    "Datum",
    "Annotation",
    "GroundTruth",
    "Prediction",
    "Filter",
    "Label",
]



import json
import logging
import math
import os
import time
from dataclasses import asdict, dataclass
from typing import Callable, List, Optional, TypeVar, Union
from urllib.parse import urlencode, urljoin

import requests
from packaging import version

from valor import __version__ as client_version
from valor.enums import TableStatus
from valor.exceptions import (
    ClientAlreadyConnectedError,
    ClientConnectionFailed,
    ClientException,
    ClientNotConnectedError,
)
from valor.schemas import EvaluationRequest

T = TypeVar("T")


def wait_for_predicate(
    update_func: Callable[[], T],
    pred: Callable[[T], bool],
    timeout: Optional[int],
    interval: float = 1.0,
) -> T:
    """Waits for a condition to become true.

    Repeatedly calls `update_func` to retrieve a new value and checks if the
    condition `pred` is satisfied.  If `pred` is not satisfied within `timeout`
    seconds, raises a TimeoutError.  Polls every `interval` seconds.

    Parameters
    ----------
    update_func:
        A callable that returns a value of type T.
    pred:
        A predicate callable that takes an argument of type T and returns a boolean.
    timeout:
        The maximum number of seconds to wait for the condition to become
        true. If None, waits indefinitely.
    interval:
        The time in seconds between consecutive calls to `update_func`.

    Returns
    ------
    T
        The final value for which `pred` returned True.

    Raises
    ----------
    TimeoutError
        If the condition is not met within `timeout` seconds.

    """
    t_start = time.time()
    state = update_func()
    while not pred(state):
        time.sleep(interval)
        if timeout and time.time() - t_start > timeout:
            raise TimeoutError
        state = update_func()
    return state


def get_json_size(json_obj: object, encoding: str = "utf-8") -> int:
    """
    Returns the number of bytes to describe the json as a string.

    Parameters
    ----------
    json_obj : object
        A JSON-compatible object.
    encoding : str, default='utf-8'
        The method used to encode the string object into a bytes format.

    Returns
    -------
    int
        The size of the object in bytes.
    """
    return len(
        json.dumps(json_obj, ensure_ascii=False).encode(
            encoding
        )  # this outputs bytes
    )


def _chunk_list(
    json_list: list, chunk_size_bytes: int, encoding: str = "utf-8"
) -> List[list]:
    """
    Chunks a list into smaller lists.

    Parameters
    ----------
    json_list : list
        A list of JSON-compatible objects.
    chunk_size_bytes : int
        The maximum number of bytes that a multi-element array can use.
    encoding : str, default='utf-8'
        The method used to encode the string object into a bytes format.

    Returns
    -------
    List[list]
        A list of lists containing JSON-compatible objects.
    """
    # edge case
    if len(json_list) == 1:
        if get_json_size(json_list, encoding) > chunk_size_bytes:
            logging.warning(
                f"Attempting to POST an object that is larger than {chunk_size_bytes} bytes."
            )
        return [json_list]

    n_elements = len(json_list)
    avg_element_size = get_json_size(json_list) / n_elements
    n_elements_per_chunk = int(chunk_size_bytes / avg_element_size)

    # return a list of individual elements if n_elements_per_chunk is zero
    if n_elements_per_chunk == 0:
        return [[element] for element in json_list]

    n_chunks = math.ceil(n_elements / n_elements_per_chunk)

    chunks = []
    for i in range(n_chunks):
        lhi = i * n_elements_per_chunk
        rhi = (i + 1) * n_elements_per_chunk
        new_chunk_size = get_json_size(
            json_obj=json_list[lhi:rhi], encoding=encoding
        )
        if (
            new_chunk_size > chunk_size_bytes
        ):  # Recursively chunk if still too large.
            chunks.extend(
                _chunk_list(
                    json_list=json_list[lhi:rhi],
                    chunk_size_bytes=chunk_size_bytes,
                    encoding=encoding,
                )
            )
        else:
            chunks.append(json_list[lhi:rhi])
    return chunks


@dataclass
class ClientConnection:
    """
    Valor client object for interacting with the api.

    Parameters
    ----------
    host : str
        The host to connect to. Should start with "http://" or "https://".
    access_token : str
        The access token for the host (if the host requires authentication).

    Raises
    ------
        ClientConnectionFailed:
            If a connection could not be established.
    """

    host: str
    access_token: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None

    def __post_init__(self):

        if not (
            self.host.startswith("http://") or self.host.startswith("https://")
        ):
            raise ValueError(
                f"host must stat with 'http://' or 'https://' but got {self.host}"
            )

        if not self.host.endswith("/"):
            self.host += "/"
        self.access_token = os.getenv("VALOR_ACCESS_TOKEN", self.access_token)
        self.username = self.username or os.getenv("VALOR_USERNAME")
        self.password = self.password or os.getenv("VALOR_PASSWORD")

        if self.username and self.password and self.access_token:
            raise ValueError(
                "You can only provide either a username and password or an access token, not both."
            )

        if self.username and self.password:
            self._using_username_password = True
            self._get_access_token_from_username_and_password()
        else:
            self._using_username_password = False

        # check the connection by getting the api version number
        try:
            api_version = self.get_api_version()
        except Exception as e:
            raise ClientConnectionFailed(str(e))

        self._validate_version(
            client_version=client_version, api_version=api_version
        )

        success_str = f"Successfully connected to host at {self.host}"
        print(success_str)

    def _get_access_token_from_username_and_password(self) -> None:
        """Sets the access token from the username and password."""
        resp = self._requests_post_rel_host(
            "token",
            ignore_auth=True,
            data={"username": self.username, "password": self.password},
        )
        if resp.ok:
            self.access_token = resp.json()

    def _validate_version(self, client_version: str, api_version: str):
        """Log and/or warn users if the Valor Python client version differs from the API version."""

        def _msg(state):
            return (
                f"The Valor client version ({client_version}) is {state} than the Valor API version {api_version}"
                f"\t==========================================================================================\n"
                f"\t== Running with a mismatched client != API version may have unexpected results.\n"
                f"\t== Please update your client to \033[1;{api_version}\033[0;31m to avoid aberrant behavior.\n"
                f"\t==========================================================================================\n"
                f"\033[0m"
            )

        if not api_version:
            logging.warning("The Valor API didn't return a version number.")
        elif not client_version:
            logging.warning("The Valor client isn't versioned.")
        elif api_version == client_version:
            logging.debug(
                f"The Valor API version {api_version} matches client version {client_version}."
            )
        elif version.parse(api_version) < version.parse(client_version):
            logging.warning(_msg("newer"))
        else:
            logging.warning(_msg("older"))

    def _requests_wrapper(
        self,
        method_name: str,
        endpoint: str,
        ignore_auth: bool = False,
        *args,
        **kwargs,
    ):
        """
        Wrapper for handling API requests.

        Parameters
        ----------
        method_name : str
            The name of the method to use for the request.
        endpoint : str
            The endpoint to send the request to.
        ignore_auth : bool, default=False
            Option to ignore authentication when you know the endpoint does not
            require a bearer token. this is used by the `_get_access_token_from_username_and_password`
            to avoid infinite recursion.
        """
        accepted_methods = ["get", "post", "put", "delete"]
        if method_name not in accepted_methods:
            raise ValueError(
                f"method_name should be one of {accepted_methods}"
            )

        if endpoint[0] == "/":
            raise ValueError(
                "`endpoint` should not start with a forward slash."
            )

        url = urljoin(self.host, endpoint)
        requests_method = getattr(requests, method_name)

        tried = False
        while True:
            if self.access_token is not None:
                headers = {"Authorization": f"Bearer {self.access_token}"}
            else:
                headers = None
            resp = requests_method(url, headers=headers, *args, **kwargs)
            if not resp.ok:
                # check if unauthorized and if using username and password, get a new
                # token and try the request again
                if (
                    resp.status_code in [401, 403]
                    and self._using_username_password
                    and not tried
                    and not ignore_auth
                ):
                    self._get_access_token_from_username_and_password()
                else:
                    try:
                        raise ClientException(resp)
                    except (requests.exceptions.JSONDecodeError, KeyError):
                        resp.raise_for_status()
            else:
                break
            tried = True

        return resp

    def _requests_post_rel_host(self, endpoint: str, *args, **kwargs):
        """
        Helper for handling POST requests.
        """
        return self._requests_wrapper(
            method_name="post", endpoint=endpoint, *args, **kwargs
        )

    def _requests_get_rel_host(self, endpoint: str, *args, **kwargs):
        """
        Helper for handling GET requests.
        """
        return self._requests_wrapper(
            method_name="get", endpoint=endpoint, *args, **kwargs
        )

    def _requests_put_rel_host(self, endpoint: str, *args, **kwargs):
        """
        Helper for handling PUT requests.
        """
        return self._requests_wrapper(
            method_name="put", endpoint=endpoint, *args, **kwargs
        )

    def _requests_delete_rel_host(self, endpoint: str, *args, **kwargs):
        """
        Helper for handling DELETE requests.
        """
        return self._requests_wrapper(
            method_name="delete", endpoint=endpoint, *args, **kwargs
        )

    def create_groundtruths(
        self,
        groundtruths: List[dict],
        chunk_size_bytes: int = int(1e7),
    ) -> None:
        """
        Creates ground truths.

        `CREATE` endpoint.

        Parameters
        ----------
        groundtruths : List[dict]
            The ground truths to be created.
        chunk_size_bytes : int, default=1e7
            Maximum size of a POST'ed json in bytes. Defaults to 10MB.
        """
        chunked_groundtruths = _chunk_list(
            json_list=groundtruths,
            chunk_size_bytes=chunk_size_bytes,
        )
        for chunk in chunked_groundtruths:
            self._requests_post_rel_host(
                "groundtruths",
                json=chunk,
            )

    def get_groundtruth(
        self,
        dataset_name: str,
        datum_uid: str,
    ) -> dict:
        """
        Get a particular ground truth.

        `GET` endpoint.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset the datum belongs to.
        datum_uid : str
            The uid of the desired datum.

        Returns
        ----------
        dict
            The requested ground truth.
        """
        return self._requests_get_rel_host(
            f"groundtruths/dataset/{dataset_name}/datum/{datum_uid}"
        ).json()

    def create_predictions(
        self,
        predictions: List[dict],
        chunk_size_bytes: int = int(1e7),
    ) -> None:
        """
        Creates predictions.

        `CREATE` endpoint.

        Parameters
        ----------
        predictions : List[dict]
            The predictions to be created.
        chunk_size_bytes : int, default=1e7
            Maximum size of a POST'ed json in bytes. Defaults to 10MB.
        """
        chunked_predictions = _chunk_list(
            json_list=predictions,
            chunk_size_bytes=chunk_size_bytes,
        )
        for chunk in chunked_predictions:
            self._requests_post_rel_host(
                "predictions",
                json=chunk,
            )

    def get_prediction(
        self,
        dataset_name: str,
        model_name: str,
        datum_uid: str,
    ) -> dict:
        """
        Get a particular prediction.

        `GET` endpoint.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset the datum belongs to.
        model_name : str
            The name of the model that made the prediction.
        datum_uid : str
            The uid of the desired datum.

        Returns
        ----------
        dict
            The requested prediction.
        """
        return self._requests_get_rel_host(
            f"predictions/model/{model_name}/dataset/{dataset_name}/datum/{datum_uid}",
        ).json()

    def get_labels(
        self,
        filter_: Optional[dict] = None,
    ) -> List[dict]:
        """
        Gets all labels using an optional filter.

        `GET` endpoint.
        """
        kwargs = {}
        if filter_:
            kwargs["params"] = {k: json.dumps(v) for k, v in filter_.items()}
        return self._requests_get_rel_host("labels", **kwargs).json()

    def get_labels_from_dataset(self, name: str) -> List[dict]:
        """
        Get all labels associated with a dataset's ground truths.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset to search by.

        Returns
        ------
        List[dict]
            A list of labels.
        """
        return self._requests_get_rel_host(f"labels/dataset/{name}").json()

    def get_labels_from_model(self, name: str) -> List[dict]:
        """
        Get all labels associated with a model's predictions.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the model to search by.

        Returns
        ------
        List[dict]
            A list of labels.
        """
        return self._requests_get_rel_host(f"labels/model/{name}").json()

    def create_dataset(self, dataset: dict):
        """
        Creates a dataset.

        `CREATE` endpoint.

        Parameters
        ----------
        dataset : dict
            A dictionary describing dataset attributes. See `valor.coretypes.Dataset` for reference.
        """
        self._requests_post_rel_host("datasets", json=dataset)

    def get_datasets(self, filter_: Optional[dict] = None) -> List[dict]:
        """
        Get all datasets with option to filter.

        `GET` endpoint.

        Parameters
        ----------
        filter_ : Filter, optional
            Optional filter to constrain by.

        Returns
        ------
        List[dict]
            A list of dictionaries describing all the datasets attributed to the `Client` object.
        """
        kwargs = {}
        if filter_:
            kwargs["params"] = {k: json.dumps(v) for k, v in filter_.items()}
        return self._requests_get_rel_host("datasets", **kwargs).json()

    def get_dataset(self, name: str) -> dict:
        """
        Gets a dataset by name.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset to fetch.

        Returns
        -------
        dict
            A dictionary containing all of the associated dataset attributes.
        """
        return self._requests_get_rel_host(f"datasets/{name}").json()

    def get_dataset_status(self, name: str) -> TableStatus:
        """
        Get the state of a given dataset.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset we want to fetch the state of.

        Returns
        ------
        TableStatus
            The state of the dataset.
        """
        resp = self._requests_get_rel_host(f"datasets/{name}/status").json()
        return TableStatus(resp)

    def get_dataset_summary(self, name: str) -> dict:
        """
        Gets the summary of a dataset.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset to create a summary for.

        Returns
        -------
        dict
            A dictionary containing the dataset summary.
        """
        return self._requests_get_rel_host(f"datasets/{name}/summary").json()

    def finalize_dataset(self, name: str) -> None:
        """
        Finalizes a dataset such that new ground truths cannot be added to it.

        `PUT` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset.
        """
        return self._requests_put_rel_host(f"datasets/{name}/finalize")

    def delete_dataset(self, name: str) -> None:
        """
        Deletes a dataset.

        `DELETE` endpoint.

        Parameters
        ----------
        name : str
            The name of the dataset to be deleted.
        """
        self._requests_delete_rel_host(f"datasets/{name}")

    def get_datums(self, filter_: Optional[dict] = None) -> List[dict]:
        """
        Get all datums using an optional filter.

        `GET` endpoint.

        Parameters
        ----------
        filter_ : dict, optional
            Optional filter to constrain by.

        Returns
        -------
        List[dict]
            A list of dictionaries describing all the datums of the specified dataset.
        """
        kwargs = {}
        if filter_:
            kwargs["params"] = {k: json.dumps(v) for k, v in filter_.items()}
        return self._requests_get_rel_host("data", **kwargs).json()

    def get_datum(
        self,
        dataset_name: str,
        uid: str,
    ) -> dict:
        """
        Get datum.
        `GET` endpoint.
        Parameters
        ----------
        dataset_name : str
            The dataset the datum belongs to.
        uid : str
            The UID of the datum.
        Returns
        -------
        dict
            A dictionary describing a datum.
        """
        return self._requests_get_rel_host(
            f"/data/dataset/{dataset_name}/uid/{uid}"
        ).json()

    def create_model(self, model: dict) -> None:
        """
        Creates a model.

        `CREATE` endpoint.

        Parameters
        ----------
        model : dict
            A dictionary describing model attributes. See `valor.coretypes.Model` for reference.
        """
        self._requests_post_rel_host("models", json=model)

    def get_models(self, filter_: Optional[dict] = None) -> List[dict]:
        """
        Get all models using an optional filter.

        `GET` endpoint.

        Parameters
        ----------
        filter_ : Filter, optional
            Optional filter to constrain by.

        Returns
        ------
        List[dict]
            A list of dictionaries describing all the models.
        """
        kwargs = {}
        if filter_:
            kwargs["params"] = {k: json.dumps(v) for k, v in filter_.items()}
        return self._requests_get_rel_host("models", **kwargs).json()

    def get_model(self, name: str) -> dict:
        """
        Gets a model by name.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the model to fetch.

        Returns
        -------
        dict
            A dictionary containing all of the associated model attributes.
        """
        return self._requests_get_rel_host(f"models/{name}").json()

    def get_model_eval_requests(self, name: str) -> List[dict]:
        """
        Get all evaluations that have been created for a model.

        This does not return evaluation results.

        `GET` endpoint.

        Parameters
        ----------
        name : str
            The name of the model.

        Returns
        -------
        List[dict]
            A list of evaluations.
        """
        return self._requests_get_rel_host(
            f"/models/{name}/eval-requests"
        ).json()

    def get_model_status(
        self,
        dataset_name: str,
        model_name: str,
    ) -> TableStatus:
        """
        Get the state of a given model over a dataset.

        `GET` endpoint.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset that the model is operating over.
        model_name : str
            The name of the model we want to fetch the state of.

        Returns
        ------
        TableStatus
            The state of the `Model`.
        """
        resp = self._requests_get_rel_host(
            f"models/{model_name}/dataset/{dataset_name}/status"
        ).json()
        return TableStatus(resp)

    def finalize_inferences(
        self,
        dataset_name: str,
        model_name: str,
    ) -> None:
        """
        Finalizes a model-dataset pairing such that new predictions cannot be added to it.

        `PUT` endpoint.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset.
        model_name : str
            The name of the model.
        """
        return self._requests_put_rel_host(
            f"models/{model_name}/datasets/{dataset_name}/finalize"
        ).json()

    def delete_model(self, name: str) -> None:
        """
        Deletes a model.

        `DELETE` endpoint.

        Parameters
        ----------
        name : str
            The name of the model to be deleted.
        """
        self._requests_delete_rel_host(f"models/{name}")

    def evaluate(self, request: EvaluationRequest) -> List[dict]:
        """
        Creates as many evaluations as necessary to fulfill the request.

        `CREATE` endpoint.

        Parameters
        ----------
        request : schemas.EvaluationRequest
            The requested evaluation parameters.

        Returns
        -------
        List[dict]
            A list of evaluations that meet the parameters.
        """
        return self._requests_post_rel_host(
            "evaluations", json=asdict(request)
        ).json()

    def get_evaluations(
        self,
        *,
        evaluation_ids: Optional[List[int]] = None,
        models: Optional[List[str]] = None,
        datasets: Optional[List[str]] = None,
    ) -> List[dict]:
        """
        Returns all evaluations associated with user-supplied dataset and/or model names.

        `GET` endpoint.

        Parameters
        ----------
        evaluation_ids : List[int], optional
            A list of job IDs to return metrics for.
        models : List[str], optional
            A list of model names that we want to return metrics for.
        datasets : List[str], optional
            A list of dataset names that we want to return metrics for.

        Returns
        -------
        List[dict]
            List of dictionaries describing the returned evaluations.
        """

        if not (evaluation_ids or models or datasets):
            raise ValueError(
                "Please provide at least one evaluation_id, model name, or dataset name"
            )

        def _build_query_param(param_name, element, typ):
            """Parse `element` to a list of `typ`, return a dict that can be urlencoded."""
            if not element:
                return {}
            if isinstance(element, typ):
                element = [element]
            return {param_name: ",".join(map(str, element))}

        params = {
            **_build_query_param("evaluation_ids", evaluation_ids, int),
            **_build_query_param("models", models, str),
            **_build_query_param("datasets", datasets, str),
        }

        query_str = urlencode(params)
        endpoint = f"evaluations?{query_str}"

        return self._requests_get_rel_host(endpoint).json()

    def get_user(self) -> Union[str, None]:
        """
        Gets the users e-mail address (in the case when auth is enabled)
        or returns None in the case of a no-auth back end.

        `GET` endpoint.

        Returns
        -------
        Union[str, None]
            The user's email address or `None` if it doesn't exist.
        """
        resp = self._requests_get_rel_host("user").json()
        return resp["email"]

    def get_api_version(self) -> str:
        """
        Gets the version number of the API.

        `GET` endpoint.

        Returns
        -------
        Union[str, None]
            The api version or `None` if it doesn't exist.
        """
        resp = self._requests_get_rel_host("api-version").json()
        return resp["api_version"]

    def health(self) -> str:
        """
        Checks if service is healthy.

        `GET` endpoint.
        """
        resp = self._requests_get_rel_host("user").json()
        return resp["status"]

    def ready(self) -> str:
        """
        Checks if service is ready.

        `GET` endpoint.
        """
        resp = self._requests_get_rel_host("user").json()
        return resp["status"]


def _create_connection():
    """
    Creates and manages a connection to the Valor API.

    This function initializes a connection closure that can be used to establish and retrieve a client connection to the Valor API. It returns two functions: `connect` and `get_connection`.

    The `connect` function is used to establish a new connection to the API, either with a new host or by reconnecting to an existing host. It raises an error if a connection is already established and `reconnect` is not set to `True`.

    The `get_connection` function is used to retrieve the current active connection. It raises an error if there's no active connection.

    Returns
    -------
    tuple
        (connect, get_connection)
    """
    _connection = None

    def connect(
        host: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        access_token: Optional[str] = None,
        reconnect: bool = False,
    ):
        """
        Establishes a connection to the Valor API.

        Parameters
        ----------
        host : str
            The host to connect to. Should start with "http://" or "https://".
        username: str
            The username for the host (if the host requires authentication).
        password: str
            The password for the host (if the host requires authentication).
        access_token : str
            The access token for the host (if the host requires authentication). Mutually
            exclusive with `username` and `password`.

        Raises
        ------
        ClientAlreadyConnectedError:
            If the connection has previously been established.
        ClientConnectionFailed:
            If a connection could not be established.
        """

        nonlocal _connection
        if _connection is not None and not reconnect:
            raise ClientAlreadyConnectedError
        _connection = ClientConnection(
            host,
            username=username,
            password=password,
            access_token=access_token,
        )

    def get_connection():
        """
        Gets the active client connection.

        Returns
        -------
        ClientConnection
            The active client connection.

        Raises
        ------
        ClientNotConnectedError
            If there is no active connection.
        """
        if _connection is None:
            raise ClientNotConnectedError
        return _connection

    def reset_connection():
        """
        Resets the connection to its initial state.
        """
        nonlocal _connection
        _connection = None

    return connect, get_connection, reset_connection


connect, get_connection, reset_connection = _create_connection()



import pytest

from valor.schemas import List as SymbolicList
from valor.schemas.symbolic.collections import StaticCollection
from valor.schemas.symbolic.types import Bool, Float, Integer, String


def test_static_collection_init():
    class A(StaticCollection):
        w: Integer
        x: Float
        y: String
        z: Bool

    # test that kwargs are required
    with pytest.raises(ValueError):
        A()


def test_static_collection_symbol():
    class A(StaticCollection):
        w: Integer
        x: Float
        y: String
        z: Bool

    # test that the 'symbolic' classmethod is the same as passing a symbol
    symA = A.symbolic()
    assert symA.to_dict() == A.symbolic(name="a").to_dict()

    # test symbolic usage
    assert symA.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": None,
            "name": "a",
            "key": None,
            "attribute": None,
        },
    }

    # test that members are also symbolic
    assert symA.w.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": "a",
            "name": "w",
            "key": None,
            "attribute": None,
        },
    }
    assert symA.x.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": "a",
            "name": "x",
            "key": None,
            "attribute": None,
        },
    }
    assert symA.y.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": "a",
            "name": "y",
            "key": None,
            "attribute": None,
        },
    }
    assert symA.z.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": "a",
            "name": "z",
            "key": None,
            "attribute": None,
        },
    }


def test_static_collection_value():
    class A(StaticCollection):
        w: Integer
        x: Float
        y: String
        z: Bool

    encoding = {"w": 101, "x": 0.123, "y": "foobar", "z": True}

    # test that casting to symbolics is implicit
    v1 = A(w=101, x=0.123, y="foobar", z=True)
    v2 = A(w=Integer(101), x=Float(0.123), y=String("foobar"), z=Bool(True))
    v3 = A(w=101, x=Float(0.123), y=String("foobar"), z=True)
    assert v1.to_dict() == v2.to_dict()
    assert v1.to_dict() == v3.to_dict()

    # test that kwargs can be loaded by dictionary
    v4 = A(**encoding)
    v5 = A(**encoding)
    assert v1.to_dict() == v4.to_dict()
    assert v1.to_dict() == v5.to_dict()

    # test dictionary generation
    assert v1.to_dict() == {
        "type": "a",
        "value": {"w": 101, "x": 0.123, "y": "foobar", "z": True},
    }

    # test value members
    assert v1.w.to_dict() == {"type": "integer", "value": 101}
    assert v1.x.to_dict() == {"type": "float", "value": 0.123}
    assert v1.y.to_dict() == {"type": "string", "value": "foobar"}
    assert v1.z.to_dict() == {"type": "bool", "value": True}


def test__get_static_types():
    class A(StaticCollection):
        w: Integer
        x: "Float"
        y: "String"
        z: Bool

    # test parsing of forward references
    assert A._get_static_types() == {
        "w": Integer,
        "x": Float,
        "y": String,
        "z": Bool,
    }

    # test lists of variables (note: these are not directly comparable)
    class B(StaticCollection):
        w: SymbolicList[Integer]
        x: SymbolicList[Float]
        y: SymbolicList[String]
        z: SymbolicList[Bool]

    types_ = B._get_static_types()
    assert types_["w"].get_element_type() == Integer
    assert types_["x"].get_element_type() == Float
    assert types_["y"].get_element_type() == String
    assert types_["z"].get_element_type() == Bool



class ClientException(Exception):
    def __init__(self, resp):
        self.status_code = resp.status_code
        self.detail = resp.json()["detail"]
        super().__init__(str(self.detail))


class ClientAlreadyConnectedError(Exception):
    def __init__(self):
        super().__init__("Client already connected.")


class ClientNotConnectedError(Exception):
    def __init__(self):
        super().__init__("Client not connected.")


class ClientConnectionFailed(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)



from typing import Tuple

import pytest

from valor.schemas.symbolic.operators import (
    And,
    AppendableFunction,
    Function,
    Negate,
    OneArgumentFunction,
    Or,
    TwoArgumentFunction,
    Xor,
)
from valor.schemas.symbolic.types import Float, Integer, String


@pytest.fixture
def variables() -> Tuple[Integer, String, Float]:
    x = Integer(1)
    y = String("2")
    z = Float(0.3)
    return (x, y, z)


def test_function(variables):
    x, y, z = variables

    # test stringify
    assert Function(x, y, z).__repr__() == "Function(1, '2', 0.3)"
    assert Function(x, y, z).__str__() == "Function(1, '2', 0.3)"

    # test dictionary generation
    assert Function(x, y, z).to_dict() == {
        "op": "function",
        "args": [
            {"type": "integer", "value": 1},
            {"type": "string", "value": "2"},
            {"type": "float", "value": 0.3},
        ],
    }

    # test stringify w/ operator
    assert issubclass(And, Function)
    assert And._operator is not None
    assert And(x, y, z).__repr__() == "And(1, '2', 0.3)"
    assert And(x, y, z).__str__() == "(1 & '2' & 0.3)"

    # test logical operators
    assert type(Function(x) & Function(y)) is And
    assert type(Function(x) | Function(y)) is Or
    assert type(Function(x) ^ Function(y)) is Xor
    assert type(~Function(x)) is Negate

    # test requirement that args must have a 'to_dict' method.
    with pytest.raises(ValueError):
        Function(1)
    with pytest.raises(ValueError):
        Function("2")
    with pytest.raises(ValueError):
        Function(0.3)


def test_appendable_function(variables):
    assert issubclass(AppendableFunction, Function)

    x, y, z = variables

    # test case where too few args
    with pytest.raises(TypeError):
        AppendableFunction(x)  # type: ignore - edge case test

    # test that all appendable functions define a overloadable function
    assert issubclass(And, AppendableFunction)
    assert issubclass(Or, AppendableFunction)
    assert issubclass(Xor, AppendableFunction)

    # test append
    f = AppendableFunction(x, y)
    f.append(z)
    assert f.to_dict() == {
        "op": "appendablefunction",
        "args": [
            {"type": "integer", "value": 1},
            {"type": "string", "value": "2"},
            {"type": "float", "value": 0.3},
        ],
    }

    # continue append on the subclass 'And'
    f1 = And(x, y)
    f1.append(z)
    assert f1.to_dict() == {
        "op": "and",
        "args": [
            {"type": "integer", "value": 1},
            {"type": "string", "value": "2"},
            {"type": "float", "value": 0.3},
        ],
    }
    assert f1.to_dict() == And(x, y, z).to_dict()
    assert f1.__repr__() == And(x, y, z).__repr__()
    assert f1.__str__() == And(x, y, z).__str__()

    # test that nested AND's collapse into one
    f2 = And(And(x, y), z)
    assert f1.to_dict() == f2.to_dict()

    # test '&' operator overload
    e1 = Integer.symbolic() == x
    e2 = String.symbolic() == y
    e3 = Float.symbolic() == z
    f3 = e1 & e2 & e3
    f4 = e1 & (e2 & e3)
    f5 = (e1 & e2) & e3
    assert f3.to_dict() == f4.to_dict()
    assert f3.to_dict() == f5.to_dict()
    assert f3.to_dict() == {
        "op": "and",
        "args": [
            {
                "op": "eq",
                "lhs": {
                    "type": "symbol",
                    "value": {
                        "owner": None,
                        "name": "integer",
                        "key": None,
                        "attribute": None,
                    },
                },
                "rhs": {"type": "integer", "value": 1},
            },
            {
                "op": "eq",
                "lhs": {
                    "type": "symbol",
                    "value": {
                        "owner": None,
                        "name": "string",
                        "key": None,
                        "attribute": None,
                    },
                },
                "rhs": {"type": "string", "value": "2"},
            },
            {
                "op": "eq",
                "lhs": {
                    "type": "symbol",
                    "value": {
                        "owner": None,
                        "name": "float",
                        "key": None,
                        "attribute": None,
                    },
                },
                "rhs": {"type": "float", "value": 0.3},
            },
        ],
    }


def test_one_arg_function(variables):
    assert issubclass(OneArgumentFunction, Function)

    x, _, _ = variables
    f = OneArgumentFunction(x)

    # test dictionary generation
    assert f.to_dict() == {
        "op": "oneargumentfunction",
        "arg": {"type": "integer", "value": 1},
    }


def test_two_arg_function(variables):
    assert issubclass(TwoArgumentFunction, Function)

    x, y, z = variables
    f = TwoArgumentFunction(x, y)

    # test memebers
    assert f.lhs == x
    assert f.rhs == y

    # test dictionary generation
    assert f.to_dict() == {
        "op": "twoargumentfunction",
        "lhs": {"type": "integer", "value": 1},
        "rhs": {"type": "string", "value": "2"},
    }

    # test cases where too few args are provided
    with pytest.raises(TypeError):
        TwoArgumentFunction(x)  # type: ignore - edge case test
    # test case where too many args are provided
    with pytest.raises(TypeError):
        TwoArgumentFunction(x, y, z)  # type: ignore - edge case test



from typing import Any, Dict, List, Optional, Union

import numpy as np

from valor.enums import TaskType
from valor.schemas.symbolic.types import (
    Box,
    Dictionary,
    Embedding,
    Equatable,
    Float,
)
from valor.schemas.symbolic.types import List as SymbolicList
from valor.schemas.symbolic.types import (
    Polygon,
    Raster,
    String,
    TaskTypeEnum,
    Variable,
    get_type_by_name,
)


def _get_schema_type_by_name(name: str):
    types_ = {
        "label": Label,
        "annotation": Annotation,
        "datum": Datum,
    }
    return get_type_by_name(name=name, additional_types=types_)


class StaticCollection(Equatable):
    """
    A static collection is a Variable that defines its contents by static attributes.
    """

    def __init__(self, **kwargs):
        if set(kwargs.keys()) != set(self._get_static_types().keys()):
            kwarg_keys = set(kwargs.keys())
            static_keys = set(self._get_static_types().keys())
            raise ValueError(
                f"Expected the following keyword arguments '{static_keys}'. Received '{kwarg_keys}'."
            )
        for k, v in kwargs.items():
            setattr(self, k, v)
        super().__init__(value=None)

    @classmethod
    def nullable(cls, *args, **kwargs):
        """
        Initializes variable with an optional value.
        """
        raise NotImplementedError(
            "Static collections do not define 'nullable'."
        )

    @classmethod
    def symbolic(
        cls,
        name: Optional[str] = None,
        key: Optional[str] = None,
        attribute: Optional[str] = None,
        owner: Optional[str] = None,
    ):
        """
        Initializes the object and its attributes as symbols.

        Parameters
        ----------
        name : str, optional
            The name of the symbol.
        key : str, optional
            The key of the value if its a dictionary element.
        attribute : str, optional
            The name of a an attribute this symbol represents.
        owner : str, optional
            The name of an object that this symbol belongs to.
        """
        obj = super().symbolic(name, key, attribute, owner)
        for __name, __type in obj._get_static_types().items():
            if not issubclass(__type, Variable):
                raise TypeError
            setattr(
                obj,
                __name,
                __type.symbolic(owner=cls.__name__.lower(), name=__name),
            )
        return obj

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return dict()

    def format(self, __name: str, __value: Any) -> Any:
        """Either formats or passes throught a name-value pair."""
        if __name in self._get_static_types():
            __type = self._get_static_types()[__name]
            if not isinstance(__value, __type):
                __fmt = (
                    self.formatting()[__name]
                    if __name in self.formatting()
                    else __type
                )
                if issubclass(__type, StaticCollection):
                    return __fmt(**__value)
                else:
                    return __fmt(__value)
        return __value

    def __setattr__(self, __name: str, __value: Any) -> None:
        super().__setattr__(__name, self.format(__name, __value))

    @classmethod
    def __validate__(cls, value: Any):
        """Validate typing."""
        if value is not None:
            raise TypeError(
                "A StaticCollection does not store an internal value."
            )

    @classmethod
    def decode_value(cls, value: dict):
        """Decode object from JSON compatible dictionary."""
        return cls(**value)

    def encode_value(self):
        """Encode object to JSON compatible dictionary."""
        return {
            k: v.encode_value() for k, v in self._get_dynamic_values().items()
        }

    @classmethod
    def _get_static_types(cls) -> Dict[str, type]:
        """Returns any static members that inherit from 'Variable'."""
        fields = getattr(cls, "__annotations__", dict())
        retval = dict()
        for k, v in fields.items():
            if isinstance(v, type) and issubclass(v, Variable):
                retval[k] = v
            elif isinstance(v, str):
                retval[k] = _get_schema_type_by_name(v)
            else:
                raise NotImplementedError(
                    f"Unknown typing. Attribute '{k}' with type '{v}'."
                )
        return retval

    def _get_dynamic_values(self):
        """Returns the values of attributes that inherit from 'Variable'."""
        return {
            name: self.__getattribute__(name)
            for name in self._get_static_types().keys()
        }

    def __repr__(self):
        if self.is_symbolic:
            return super().__repr__()
        return self.encode_value().__repr__()

    def __str__(self):
        if self.is_symbolic:
            return super().__str__()
        return str(self.encode_value())

    @property
    def is_symbolic(self) -> bool:
        return super().is_symbolic


class Label(StaticCollection):
    """
    An object for labeling datasets, models, and annotations.

    Attributes
    ----------
    key : String
        The class label key.
    value : String
        The class label value.
    score : Score
        The label score.

    Examples
    --------
    >>> Label(key="k1", value="v1")
    >>> Label(key="k1", value="v1", score=None)
    >>> Label(key="k1", value="v1", score=0.9)
    """

    key: String = String.symbolic(owner="label", name="key")
    value: String = String.symbolic(owner="label", name="value")
    score: Float = Float.symbolic(owner="label", name="score")

    def __init__(
        self,
        *,
        key: str,
        value: str,
        score: Union[float, np.floating, None] = None,
    ):
        """
        Initializes an instance of a label.

        Attributes
        ----------
        key : str
            The class label key.
        value : str
            The class label value.
        score : float, optional
            The label score.
        """
        super().__init__(key=key, value=value, score=score)

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return {
            "score": Float.nullable,
        }

    def tuple(self):
        """
        Defines how the `Label` is turned into a tuple.

        Returns
        ----------
        tuple
            A tuple of the `Label's` arguments.
        """
        return (
            self.key.get_value(),
            self.value.get_value(),
            self.score.get_value(),
        )


class Annotation(StaticCollection):
    """
    A class used to annotate `GroundTruths` and `Predictions`.

    Attributes
    ----------
    task_type: TaskTypeEnum
        The task type associated with the `Annotation`.
    metadata: Dictionary
        A dictionary of metadata that describes the `Annotation`.
    labels: List[Label], optional
        A list of labels to use for the `Annotation`.
    bounding_box: Box
        A bounding box to assign to the `Annotation`.
    polygon: BoundingPolygon
        A polygon to assign to the `Annotation`.
    raster: Raster
        A raster to assign to the `Annotation`.
    embedding: List[float]
        An embedding, described by a list of values with type float and a maximum length of 16,000.

    Examples
    --------

    Classification
    >>> Annotation.create(
    ...     task_type=TaskType.CLASSIFICATION,
    ...     labels=[
    ...         Label(key="class", value="dog"),
    ...         Label(key="category", value="animal"),
    ...     ]
    ... )

    Object-Detection Box
    >>> annotation = Annotation.create(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...    bounding_box=box2,
    ... )

    Object-Detection Polygon
    >>> annotation = Annotation.create(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     polygon=BoundingPolygon(...),
    ... )

    Object-Detection Raster
    >>> annotation = Annotation.create(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=Raster(...),
    ... )

    Semantic-Segmentation Raster
    >>> annotation = Annotation.create(
    ...     task_type=TaskType.SEMANTIC_SEGMENTATION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     raster=Raster(...),
    ... )

    Defining all supported annotation types for a given `task_type` is allowed!
    >>> Annotation.create(
    ...     task_type=TaskType.OBJECT_DETECTION,
    ...     labels=[Label(key="k1", value="v1")],
    ...     bounding_box=Box(...),
    ...     polygon=BoundingPolygon(...),
    ...     raster=Raster(...),
    ... )
    """

    task_type: TaskTypeEnum = TaskTypeEnum.symbolic(
        owner="annotation", name="task_type"
    )
    metadata: Dictionary = Dictionary.symbolic(
        owner="annotation", name="metadata"
    )
    labels: SymbolicList[Label] = SymbolicList[Label].symbolic(
        owner="annotation", name="labels"
    )
    bounding_box: Box = Box.symbolic(owner="annotation", name="bounding_box")
    polygon: Polygon = Polygon.symbolic(owner="annotation", name="polygon")
    raster: Raster = Raster.symbolic(owner="annotation", name="raster")
    embedding: Embedding = Embedding.symbolic(
        owner="annotation", name="embedding"
    )

    def __init__(
        self,
        *,
        task_type: TaskType,
        metadata: Optional[dict] = None,
        labels: Optional[List[Label]] = None,
        bounding_box: Optional[Box] = None,
        polygon: Optional[Polygon] = None,
        raster: Optional[Raster] = None,
        embedding: Optional[Embedding] = None,
    ):
        """
        Constructs an annotation.

        Parameters
        ----------
        task_type: TaskTypeEnum
            The task type associated with the `Annotation`.
        metadata: Dict[str, Union[int, float, str, bool, datetime.datetime, datetime.date, datetime.time]]
            A dictionary of metadata that describes the `Annotation`.
        labels: List[Label]
            A list of labels to use for the `Annotation`.
        bounding_box: Box, optional
            A bounding box annotation.
        polygon: Polygon, optional
            A polygon annotation.
        raster: Raster, optional
            A raster annotation.
        embedding: List[float], optional
            An embedding, described by a list of values with type float and a maximum length of 16,000.
        """
        super().__init__(
            task_type=task_type,
            metadata=metadata if metadata else dict(),
            labels=labels if labels else list(),
            bounding_box=bounding_box,
            polygon=polygon,
            raster=raster,
            embedding=embedding,
        )

    @staticmethod
    def formatting() -> Dict[str, Any]:
        """Attribute format mapping."""
        return {
            "bounding_box": Box.nullable,
            "polygon": Polygon.nullable,
            "raster": Raster.nullable,
            "embedding": Embedding.nullable,
        }


class Datum(StaticCollection):
    """
    A class used to store information about a datum for either a 'GroundTruth' or a 'Prediction'.

    Attributes
    ----------
    uid : String
        The UID of the datum.
    metadata : Dictionary
        A dictionary of metadata that describes the datum.

    Examples
    --------
    >>> Datum(uid="uid1")
    >>> Datum(uid="uid1", metadata={})
    >>> Datum(uid="uid1", metadata={"foo": "bar", "pi": 3.14})
    """

    uid: String = String.symbolic(owner="datum", name="uid")
    metadata: Dictionary = Dictionary.symbolic(owner="datum", name="metadata")

    def __init__(
        self,
        *,
        uid: str,
        metadata: Optional[dict] = None,
    ):
        """
        Constructs a datum.

        Parameters
        ----------
        uid : str
            The UID of the datum.
        metadata : dict, optional
            A dictionary of metadata that describes the datum.
        """
        super().__init__(uid=uid, metadata=metadata if metadata else dict())

    def get_uid(self) -> str:
        """Extracts the uid from a datum instance."""
        return self.uid.get_value()



from __future__ import annotations

import datetime
import json
import time
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from valor.client import ClientConnection, connect, get_connection
from valor.enums import AnnotationType, EvaluationStatus, TableStatus, TaskType
from valor.exceptions import ClientException
from valor.schemas import (
    Annotation,
    Datum,
    Dictionary,
    EvaluationParameters,
    EvaluationRequest,
    Filter,
    Label,
)
from valor.schemas import List as SymbolicList
from valor.schemas import StaticCollection, String
from valor.schemas.compatibility import decode_api_format, encode_api_format

FilterType = Union[list, dict, Filter]  # TODO - Remove this


def _format_filter(filter_by: Optional[FilterType]) -> Filter:
    """
    Formats the various filter or constraint representations into a 'schemas.Filter' object.

    Parameters
    ----------
    filter_by : FilterType, optional
        The reference filter.

    Returns
    -------
    valor.schemas.Filter
        A properly formatted 'schemas.Filter' object.
    """
    if isinstance(filter_by, Filter):
        return filter_by
    elif isinstance(filter_by, list) or filter_by is None:
        filter_by = filter_by if filter_by else []
        return Filter.create(filter_by)
    elif isinstance(filter_by, dict):
        return Filter(**filter_by)
    else:
        raise TypeError


class GroundTruth(StaticCollection):
    """
    An object describing a ground truth (e.g., a human-drawn bounding box on an image).

    Attributes
    ----------
    datum : Datum
        The datum associated with the groundtruth.
    annotations : List[Annotation]
        The list of annotations associated with the groundtruth.

    Examples
    --------
    >>> GroundTruth(
    ...     datum=Datum(uid="uid1"),
    ...     annotations=[
    ...         Annotation(
    ...             task_type=TaskType.CLASSIFICATION,
    ...             labels=[Label(key="k1", value="v1")],
    ...         )
    ...     ]
    ... )
    """

    datum: Datum = Datum.symbolic(owner="groundtruth", name="datum")
    annotations: SymbolicList[Annotation] = SymbolicList[Annotation].symbolic(
        owner="groundtruth", name="annotations"
    )

    def __init__(
        self,
        *,
        datum: Datum,
        annotations: List[Annotation],
    ):
        """
        Creates a ground truth.

        Parameters
        ----------
        datum : Datum
            The datum that the ground truth is operating over.
        annotations : List[Annotation]
            The list of ground truth annotations.
        """
        super().__init__(datum=datum, annotations=annotations)

        for annotation in self.annotations:
            for label in annotation.labels:
                if label.score.get_value() is not None:
                    raise ValueError(
                        "GroundTruth labels should not have scores."
                    )


class Prediction(StaticCollection):
    """
    An object describing a prediction (e.g., a machine-drawn bounding box on an image).

    Attributes
    ----------
    datum : Datum
        The datum associated with the prediction.
    annotations : List[Annotation]
        The list of annotations associated with the prediction.

    Examples
    --------
    >>> Prediction(
    ...     datum=Datum(uid="uid1"),
    ...     annotations=[
    ...         Annotation(
    ...             task_type=TaskType.CLASSIFICATION,
    ...             labels=[
    ...                 Label(key="k1", value="v1", score=0.9),
    ...                 Label(key="k1", value="v1", score=0.1)
    ...             ],
    ...         )
    ...     ]
    ... )
    """

    datum: Datum = Datum.symbolic(owner="prediction", name="datum")
    annotations: SymbolicList[Annotation] = SymbolicList[Annotation].symbolic(
        owner="prediction", name="annotations"
    )

    def __init__(
        self,
        *,
        datum: Datum,
        annotations: List[Annotation],
    ):
        """
        Creates a prediction.

        Parameters
        ----------
        datum : Datum
            The datum that the prediction is operating over.
        annotations : List[Annotation]
            The list of predicted annotations.
        """
        super().__init__(datum=datum, annotations=annotations)

        # validation
        for annotation in self.annotations:
            task_type = annotation.task_type.get_value()
            if task_type in [
                TaskType.CLASSIFICATION,
                TaskType.OBJECT_DETECTION,
            ]:
                for label in annotation.labels:
                    label_score = label.score._value
                    if label_score is None:
                        raise ValueError(
                            f"For task type `{task_type}` prediction labels must have scores, but got `None`"
                        )
            if task_type == TaskType.CLASSIFICATION:

                label_keys_to_sum = {}
                for scored_label in annotation.labels:
                    label_key = scored_label.key.get_value()
                    label_score = scored_label.score.get_value()
                    if label_key not in label_keys_to_sum:
                        label_keys_to_sum[label_key] = 0.0
                    label_keys_to_sum[label_key] += label_score

                for k, total_score in label_keys_to_sum.items():
                    if abs(total_score - 1) > 1e-5:
                        raise ValueError(
                            "For each label key, prediction scores must sum to 1, but"
                            f" for label key {k} got scores summing to {total_score}."
                        )


class Evaluation:
    """
    Wraps `valor.client.Job` to provide evaluation-specifc members.
    """

    def __init__(
        self, connection: Optional[ClientConnection] = None, **kwargs
    ):
        """
        Defines important attributes of the API's `EvaluationResult`.

        Attributes
        ----------
        id : int
            The ID of the evaluation.
        model_name : str
            The name of the evaluated model.
        datum_filter : schemas.Filter
            The filter used to select the datums for evaluation.
        status : EvaluationStatus
            The status of the evaluation.
        metrics : List[dict]
            A list of metric dictionaries returned by the job.
        confusion_matrices : List[dict]
            A list of confusion matrix dictionaries returned by the job.
        """
        if not connection:
            connection = get_connection()
        self.conn = connection
        self.update(**kwargs)

    def update(
        self,
        *_,
        id: int,
        model_name: str,
        datum_filter: Filter,
        parameters: EvaluationParameters,
        status: EvaluationStatus,
        metrics: List[Dict],
        confusion_matrices: List[Dict],
        created_at: str,
        **kwargs,
    ):
        self.id = id
        self.model_name = model_name
        self.datum_filter = (
            Filter(**datum_filter)
            if isinstance(datum_filter, dict)
            else datum_filter
        )
        self.parameters = (
            EvaluationParameters(**parameters)
            if isinstance(parameters, dict)
            else parameters
        )
        self.status = EvaluationStatus(status)
        self.metrics = metrics
        self.confusion_matrices = confusion_matrices
        self.kwargs = kwargs
        self.ignored_pred_labels: Optional[List[Label]] = None
        self.missing_pred_labels: Optional[List[Label]] = None
        self.created_at = datetime.datetime.strptime(
            created_at, "%Y-%m-%dT%H:%M:%S.%fZ"
        ).replace(tzinfo=datetime.timezone.utc)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def poll(self) -> EvaluationStatus:
        """
        Poll the back end.

        Updates the evaluation with the latest state from the back end.

        Returns
        -------
        enums.EvaluationStatus
            The status of the evaluation.

        Raises
        ----------
        ClientException
            If an Evaluation with the given `evaluation_id` is not found.
        """
        response = self.conn.get_evaluations(evaluation_ids=[self.id])
        if not response:
            raise ClientException("Not Found")
        self.update(**response[0])
        return self.status

    def wait_for_completion(
        self,
        *,
        timeout: Optional[int] = None,
        interval: float = 1.0,
    ) -> EvaluationStatus:
        """
        Blocking function that waits for evaluation to finish.

        Parameters
        ----------
        timeout : int, optional
            Length of timeout in seconds.
        interval : float, default=1.0
            Polling interval in seconds.
        """
        t_start = time.time()
        while self.poll() not in [
            EvaluationStatus.DONE,
            EvaluationStatus.FAILED,
        ]:
            time.sleep(interval)
            if timeout and time.time() - t_start > timeout:
                raise TimeoutError
        return self.status

    def __str__(self) -> str:
        """Dumps the object into a JSON formatted string."""
        return json.dumps(self.to_dict(), indent=4)

    def to_dict(self) -> dict:
        """
        Defines how a `valor.Evaluation` object is serialized into a dictionary.

        Returns
        ----------
        dict
            A dictionary describing an evaluation.
        """
        return {
            "id": self.id,
            "model_name": self.model_name,
            "datum_filter": asdict(self.datum_filter),
            "parameters": asdict(self.parameters),
            "status": self.status.value,
            "metrics": self.metrics,
            "confusion_matrices": self.confusion_matrices,
            **self.kwargs,
        }

    def to_dataframe(
        self,
        stratify_by: Optional[Tuple[str, str]] = None,
    ):
        """
        Get all metrics associated with a Model and return them in a `pd.DataFrame`.

        Returns
        ----------
        pd.DataFrame
            Evaluation metrics being displayed in a `pd.DataFrame`.

        Raises
        ------
        ModuleNotFoundError
            This function requires the use of `pandas.DataFrame`.

        """
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Must have pandas installed to use `get_metric_dataframes`."
            )

        if not stratify_by:
            column_type = "evaluation"
            column_name = self.id
        else:
            column_type = stratify_by[0]
            column_name = stratify_by[1]

        metrics = [
            {**metric, column_type: column_name} for metric in self.metrics
        ]
        df = pd.DataFrame(metrics)
        for k in ["label", "parameters"]:
            df[k] = df[k].fillna("n/a")
        df["parameters"] = df["parameters"].apply(json.dumps)
        df["label"] = df["label"].apply(
            lambda x: f"{x['key']}: {x['value']}" if x != "n/a" else x
        )
        df = df.pivot(
            index=["type", "parameters", "label"], columns=[column_type]
        )
        return df


@dataclass
class DatasetSummary:
    """Dataclass for storing dataset summary information"""

    name: str
    num_datums: int
    num_annotations: int
    num_bounding_boxes: int
    num_polygons: int
    num_rasters: int
    task_types: List[TaskType]
    labels: List[Label]
    datum_metadata: List[dict]
    annotation_metadata: List[dict]

    def __post_init__(self):
        for i, tt in enumerate(self.task_types):
            if isinstance(tt, str):
                self.task_types[i] = TaskType(tt)
        for i, label in enumerate(self.labels):
            if isinstance(label, dict):
                self.labels[i] = Label(**label)


class Dataset(StaticCollection):
    """
    A class describing a given dataset.

    Attributes
    ----------
    name : String
        The name of the dataset.
    metadata : Dictionary
        A dictionary of metadata that describes the dataset.

    Examples
    --------
    >>> Dataset.create(name="dataset1")
    >>> Dataset.create(name="dataset1", metadata={})
    >>> Dataset.create(name="dataset1", metadata={"foo": "bar", "pi": 3.14})
    """

    name: String = String.symbolic(owner="dataset", name="name")
    metadata: Dictionary = Dictionary.symbolic(
        owner="dataset", name="metadata"
    )

    def __init__(
        self,
        *,
        name: str,
        metadata: Optional[dict] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Creates a local instance of a dataset.

        Use 'Dataset.create' classmethod to create a dataset with persistence.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict, optional
            A dictionary of metadata that describes the dataset.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        self.conn = connection
        super().__init__(name=name, metadata=metadata if metadata else dict())

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
        connection: Optional[ClientConnection] = None,
    ) -> Dataset:
        """
        Creates a dataset that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the dataset.
        metadata : dict, optional
            A dictionary of metadata that describes the dataset.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        dataset = cls(name=name, metadata=metadata, connection=connection)
        Client(dataset.conn).create_dataset(dataset)
        return dataset

    @classmethod
    def get(
        cls,
        name: str,
        connection: Optional[ClientConnection] = None,
    ) -> Union[Dataset, None]:
        """
        Retrieves a dataset from the back end database.

        Parameters
        ----------
        name : str
            The name of the dataset.

        Returns
        -------
        Union[valor.Dataset, None]
            The dataset or 'None' if it doesn't exist.
        """
        return Client(connection).get_dataset(name)

    def add_groundtruth(
        self,
        groundtruth: GroundTruth,
    ) -> None:
        """
        Add a ground truth to the dataset.

        Parameters
        ----------
        groundtruth : GroundTruth
            The ground truth to create.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=[groundtruth],
        )

    def add_groundtruths(
        self,
        groundtruths: List[GroundTruth],
    ) -> None:
        """
        Add multiple ground truths to the dataset.

        Parameters
        ----------
        groundtruths : List[GroundTruth]
            The ground truths to create.
        """
        Client(self.conn).create_groundtruths(
            dataset=self,
            groundtruths=groundtruths,
        )

    def get_groundtruth(
        self,
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_groundtruth(dataset=self, datum=datum)

    def get_name(self) -> str:
        return self.name.get_value()

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given dataset.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the dataset.
        """
        return Client(self.conn).get_labels_from_dataset(self)

    def get_datums(
        self, filter_by: Optional[FilterType] = None
    ) -> List[Datum]:
        """
        Get all datums associated with a given dataset.

        Parameters
        ----------
        filter_by
            Optional constraints to filter by.

        Returns
        ----------
        List[Datum]
            A list of `Datums` associated with the dataset.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)

        if filter_.get("dataset_names"):
            raise ValueError(
                "Cannot filter by dataset_names when calling `Dataset.get_datums`."
            )
        filter_["dataset_names"] = [self.get_name()]
        return Client(self.conn).get_datums(filter_by=filter_)

    def get_evaluations(
        self,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given dataset.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the dataset.
        """
        return Client(self.conn).get_evaluations(datasets=[self])

    def get_summary(self) -> DatasetSummary:
        """
        Get the summary of a given dataset.

        Returns
        -------
        DatasetSummary
            The summary of the dataset. This class has the following fields:

            name: name of the dataset

            num_datums: total number of datums in the dataset

            num_annotations: total number of labeled annotations in the dataset; if an
            object (such as a bounding box) has multiple labels, then each label is counted separately

            num_bounding_boxes: total number of bounding boxes in the dataset

            num_polygons: total number of polygons in the dataset

            num_rasters: total number of rasters in the dataset

            task_types: list of the unique task types in the dataset

            labels: list of the unique labels in the dataset

            datum_metadata: list of the unique metadata dictionaries in the dataset that are associated
            to datums

            groundtruth_annotation_metadata: list of the unique metadata dictionaries in the dataset that are
            associated to annotations
        """
        return Client(self.conn).get_dataset_summary(self.get_name())

    def finalize(
        self,
    ):
        """
        Finalizes the dataset such that new ground truths cannot be added to it.
        """
        return Client(self.conn).finalize_dataset(self)

    def delete(
        self,
        timeout: int = 0,
    ):
        """
        Delete the dataset from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_dataset(self.get_name(), timeout)


class Model(StaticCollection):
    """
    A class describing a model that was trained on a particular dataset.

    Attributes
    ----------
    name : String
        The name of the model.
    metadata : Dictionary
        A dictionary of metadata that describes the model.

    Examples
    --------
    >>> Model.create(name="model1")
    >>> Model.create(name="model1", metadata={})
    >>> Model.create(name="model1", metadata={"foo": "bar", "pi": 3.14})
    """

    name: String = String.symbolic(owner="model", name="name")
    metadata: Dictionary = Dictionary.symbolic(owner="model", name="metadata")

    def __init__(
        self,
        *,
        name: str,
        metadata: Optional[dict] = None,
        connection: Optional[ClientConnection] = None,
    ):
        """
        Creates a local instance of a model.

        Use 'Model.create' classmethod to create a model with persistence.

        Parameters
        ----------
        name : String
            The name of the model.
        metadata : Dictionary
            A dictionary of metadata that describes the model.
        connection : ClientConnection, optional
            An initialized client connection.
        symbol : Symbol, optional
            Symbol to represent a model.
        """
        self.conn = connection
        super().__init__(name=name, metadata=metadata if metadata else dict())

    @classmethod
    def create(
        cls,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
        connection: Optional[ClientConnection] = None,
        **_,
    ) -> Model:
        """
        Creates a model that persists in the back end.

        Parameters
        ----------
        name : str
            The name of the model.
        metadata : dict, optional
            A dictionary of metadata that describes the model.
        connection : ClientConnection, optional
            An initialized client connection.
        """
        model = cls(name=name, metadata=metadata, connection=connection)
        model.add_connection(connection)
        Client(connection).create_model(model)
        return model

    @classmethod
    def get(
        cls,
        name: str,
        connection: Optional[ClientConnection] = None,
    ) -> Union[Model, None]:
        """
        Retrieves a model from the back end database.

        Parameters
        ----------
        name : str
            The name of the model.
        connection : ClientConnnetion, optional
            An optional Valor client object for interacting with the API.

        Returns
        -------
        Union[valor.Model, None]
            The model or 'None' if it doesn't exist.
        """
        return Client(connection).get_model(name)

    def add_connection(self, connection: Optional[ClientConnection]):
        """
        Stores a pre-existing connection.

        Parameters
        ----------
        connection : ClientConnnetion, optional
            An optional Valor client object for interacting with the API.
        """
        self.conn = connection

    def add_prediction(
        self,
        dataset: Dataset,
        prediction: Prediction,
    ) -> None:
        """
        Add a prediction to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        prediction : valor.Prediction
            The prediction to create.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=[prediction],
        )

    def add_predictions(
        self,
        dataset: Dataset,
        predictions: List[Prediction],
    ) -> None:
        """
        Add multiple predictions to the model.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        predictions : List[valor.Prediction]
            The predictions to create.
        """
        Client(self.conn).create_predictions(
            dataset=dataset,
            model=self,
            predictions=predictions,
        )

    def get_name(self) -> str:
        return self.name.get_value()

    def get_prediction(
        self, dataset: Union[Dataset, str], datum: Union[Datum, str]
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        return Client(self.conn).get_prediction(
            dataset=dataset, model=self, datum=datum
        )

    def finalize_inferences(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes the model over a dataset such that new predictions cannot be added to it.
        """
        return Client(self.conn).finalize_inferences(
            dataset=dataset, model=self
        )

    def _format_constraints(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
    ) -> Filter:
        """Formats the 'datum_filter' for any evaluation requests."""

        # get list of dataset names
        dataset_names_from_obj = []
        if isinstance(datasets, list):
            dataset_names_from_obj = [
                dataset.get_name() for dataset in datasets
            ]
        elif isinstance(datasets, Dataset):
            dataset_names_from_obj = [datasets.get_name()]

        # create a 'schemas.Filter' object from the constraints.
        filter_ = _format_filter(filter_by)

        # reset model name
        filter_.model_names = None
        filter_.model_metadata = None

        # set dataset names
        if not filter_.dataset_names:
            filter_.dataset_names = []
        filter_.dataset_names.extend(dataset_names_from_obj)
        return filter_

    def _create_label_map(
        self,
        label_map: Optional[Dict[Label, Label]],
    ) -> Union[List[List[List[str]]], None]:
        """Convert a dictionary of label maps to a serializable list format."""
        if not label_map:
            return None

        if not isinstance(label_map, dict) or not all(
            [
                isinstance(key, Label) and isinstance(value, Label)
                for key, value in label_map.items()
            ]
        ):
            raise TypeError(
                "label_map should be a dictionary with valid Labels for both the key and value."
            )

        return_value = []
        for key, value in label_map.items():
            if not all(
                [
                    (
                        isinstance(v.key._value, str)
                        and isinstance(v.value._value, str)
                    )
                    for v in [key, value]
                ]
            ):
                raise TypeError
            return_value.append(
                [
                    [key.key._value, key.value._value],
                    [value.key._value, value.value._value],
                ]
            )
        return return_value

    def evaluate_classification(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        compute_pr_curves: bool = False,
    ) -> Evaluation:
        """
        Start a classification evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        compute_pr_curves: bool
            A boolean which determines whether we calculate precision-recall curves or not.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if not datasets and not filter_by:
            raise ValueError(
                "Evaluation requires the definition of either datasets, dataset filters or both."
            )

        # format request
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.get_name()],
            datum_filter=datum_filter,
            parameters=EvaluationParameters(
                task_type=TaskType.CLASSIFICATION,
                label_map=self._create_label_map(label_map=label_map),
                compute_pr_curves=compute_pr_curves,
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_detection(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        convert_annotations_to_type: Optional[AnnotationType] = None,
        iou_thresholds_to_compute: Optional[List[float]] = None,
        iou_thresholds_to_return: Optional[List[float]] = None,
        label_map: Optional[Dict[Label, Label]] = None,
        recall_score_threshold: float = 0,
        compute_pr_curves: bool = False,
        pr_curve_iou_threshold: float = 0.5,
    ) -> Evaluation:
        """
        Start an object-detection evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        convert_annotations_to_type : enums.AnnotationType, optional
            Forces the object detection evaluation to compute over this type.
        iou_thresholds_to_compute : List[float], optional
            Thresholds to compute mAP against.
        iou_thresholds_to_return : List[float], optional
            Thresholds to return AP for. Must be subset of `iou_thresholds_to_compute`.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.
        recall_score_threshold: float, default=0
            The confidence score threshold for use when determining whether to count a prediction as a true positive or not while calculating Average Recall.
        compute_pr_curves: bool, optional
            A boolean which determines whether we calculate precision-recall curves or not.
        pr_curve_iou_threshold: float, optional
            The IOU threshold to use when calculating precision-recall curves. Defaults to 0.5. Does nothing when compute_pr_curves is set to False or None.


        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion.
        """
        if iou_thresholds_to_compute is None:
            iou_thresholds_to_compute = [
                round(0.5 + 0.05 * i, 2) for i in range(10)
            ]
        if iou_thresholds_to_return is None:
            iou_thresholds_to_return = [0.5, 0.75]

        # format request
        parameters = EvaluationParameters(
            task_type=TaskType.OBJECT_DETECTION,
            convert_annotations_to_type=convert_annotations_to_type,
            iou_thresholds_to_compute=iou_thresholds_to_compute,
            iou_thresholds_to_return=iou_thresholds_to_return,
            label_map=self._create_label_map(label_map=label_map),
            recall_score_threshold=recall_score_threshold,
            compute_pr_curves=compute_pr_curves,
            pr_curve_iou_threshold=pr_curve_iou_threshold,
        )
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.get_name()],
            datum_filter=datum_filter,
            parameters=parameters,
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def evaluate_segmentation(
        self,
        datasets: Optional[Union[Dataset, List[Dataset]]] = None,
        filter_by: Optional[FilterType] = None,
        label_map: Optional[Dict[Label, Label]] = None,
    ) -> Evaluation:
        """
        Start a semantic-segmentation evaluation job.

        Parameters
        ----------
        datasets : Union[Dataset, List[Dataset]], optional
            The dataset or list of datasets to evaluate against.
        filter_by : FilterType, optional
            Optional set of constraints to filter evaluation by.
        label_map : Dict[Label, Label], optional
            Optional mapping of individual labels to a grouper label. Useful when you need to evaluate performance using labels that differ across datasets and models.

        Returns
        -------
        Evaluation
            A job object that can be used to track the status of the job and get the metrics of it upon completion
        """
        # format request
        datum_filter = self._format_constraints(datasets, filter_by)
        request = EvaluationRequest(
            model_names=[self.get_name()],
            datum_filter=datum_filter,
            parameters=EvaluationParameters(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                label_map=self._create_label_map(label_map=label_map),
            ),
        )

        # create evaluation
        evaluation = Client(self.conn).evaluate(request)
        if len(evaluation) != 1:
            raise RuntimeError
        return evaluation[0]

    def delete(self, timeout: int = 0):
        """
        Delete the `Model` object from the back end.

        Parameters
        ----------
        timeout : int, default=0
            Sets a timeout in seconds.
        """
        Client(self.conn).delete_model(self.get_name(), timeout)

    def get_labels(
        self,
    ) -> List[Label]:
        """
        Get all labels associated with a given model.

        Returns
        ----------
        List[Label]
            A list of `Labels` associated with the model.
        """
        return Client(self.conn).get_labels_from_model(self)

    def get_evaluations(
        self,
    ) -> List[Evaluation]:
        """
        Get all evaluations associated with a given model.

        Returns
        ----------
        List[Evaluation]
            A list of `Evaluations` associated with the model.
        """
        return Client(self.conn).get_evaluations(models=[self])


class Client:
    """
    Valor client object for interacting with the api.

    Parameters
    ----------
    connection : ClientConnection, optional
        Option to use an existing connection object.
    """

    def __init__(self, connection: Optional[ClientConnection] = None):
        if not connection:
            connection = get_connection()
        self.conn = connection

    @classmethod
    def connect(
        cls,
        host: str,
        access_token: Optional[str] = None,
        reconnect: bool = False,
    ) -> Client:
        """
        Establishes a connection to the Valor API.

        Parameters
        ----------
        host : str
            The host to connect to. Should start with "http://" or "https://".
        access_token : str
            The access token for the host (if the host requires authentication).
        """
        connect(host=host, access_token=access_token, reconnect=reconnect)
        return cls(get_connection())

    def get_labels(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Label]:
        """
        Gets all labels using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        filter_ = _format_filter(filter_by)
        filter_ = asdict(filter_)
        return [Label(**label) for label in self.conn.get_labels(filter_)]

    def get_labels_from_dataset(
        self, dataset: Union[Dataset, str]
    ) -> List[Label]:
        """
        Get all labels associated with a dataset's ground truths.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        return [
            Label(**label)
            for label in self.conn.get_labels_from_dataset(dataset_name)
        ]

    def get_labels_from_model(self, model: Union[Model, str]) -> List[Label]:
        """
        Get all labels associated with a model's ground truths.

        Parameters
        ----------
        model : valor.Model
            The model to search by.

        Returns
        ------
        List[valor.Label]
            A list of labels.
        """
        model_name = model.get_name() if isinstance(model, Model) else model
        return [
            Label(**label)
            for label in self.conn.get_labels_from_model(model_name)
        ]

    def create_dataset(
        self,
        dataset: Union[Dataset, dict],
    ) -> None:
        """
        Creates a dataset.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset to create.
        """
        if isinstance(dataset, Dataset):
            dataset = encode_api_format(dataset)
        self.conn.create_dataset(dataset)

    def create_groundtruths(
        self,
        dataset: Dataset,
        groundtruths: List[GroundTruth],
    ):
        """
        Creates ground truths.

        Parameters
        ----------

        dataset : valor.Dataset
            The dataset to create the ground truth for.
        groundtruths : List[valor.GroundTruth]
            The ground truths to create.
        """
        groundtruths_json = []
        for groundtruth in groundtruths:
            if not isinstance(groundtruth, GroundTruth):
                raise TypeError(
                    f"Expected ground truth to be of type 'valor.GroundTruth' not '{type(groundtruth)}'."
                )
            if not isinstance(groundtruth.annotations._value, list):
                raise TypeError
            groundtruth_dict = encode_api_format(groundtruth)
            groundtruth_dict["datum"]["dataset_name"] = dataset.get_name()
            groundtruths_json.append(groundtruth_dict)
        self.conn.create_groundtruths(groundtruths_json)

    def get_groundtruth(
        self,
        dataset: Union[Dataset, str],
        datum: Union[Datum, str],
    ) -> Union[GroundTruth, None]:
        """
        Get a particular ground truth.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[GroundTruth, None]
            The matching ground truth or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        datum_uid = datum.get_uid() if isinstance(datum, Datum) else datum
        try:
            resp = self.conn.get_groundtruth(
                dataset_name=dataset_name, datum_uid=datum_uid
            )
            resp = decode_api_format(resp)
            return GroundTruth(**resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def finalize_dataset(self, dataset: Union[Dataset, str]) -> None:
        """
        Finalizes a dataset such that new ground truths cannot be added to it.

        Parameters
        ----------
        dataset : str
            The dataset to be finalized.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        return self.conn.finalize_dataset(name=dataset_name)

    def get_dataset(
        self,
        name: str,
    ) -> Union[Dataset, None]:
        """
        Gets a dataset by name.

        Parameters
        ----------
        name : str
            The name of the dataset to fetch.

        Returns
        -------
        Union[Dataset, None]
            A Dataset with a matching name, or 'None' if one doesn't exist.
        """
        try:
            resp = decode_api_format(self.conn.get_dataset(name))
            dataset = Dataset(
                **resp,
                connection=self.conn,
            )
            return dataset
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_datasets(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Dataset]:
        """
        Get all datasets, with an option to filter results according to some user-defined parameters.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Dataset]
            A list of datasets.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        dataset_list = []
        for kwargs in self.conn.get_datasets(filter_):
            kwargs = decode_api_format(kwargs)
            dataset = Dataset(**kwargs, connection=self.conn)
            dataset_list.append(dataset)
        return dataset_list

    def get_datums(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Datum]:
        """
        Get all datums using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        -------
        List[valor.Datum]
            A list datums.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        return [
            Datum(**decode_api_format(datum))
            for datum in self.conn.get_datums(filter_)
        ]

    def get_datum(
        self,
        dataset: Union[Dataset, str],
        uid: str,
    ) -> Union[Datum, None]:
        """
        Get datum.
        `GET` endpoint.
        Parameters
        ----------
        dataset : valor.Dataset
            The dataset the datum belongs to.
        uid : str
            The UID of the datum.
        Returns
        -------
        valor.Datum
            The requested datum or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        try:
            resp = self.conn.get_datum(dataset_name=dataset_name, uid=uid)
            resp = decode_api_format(resp)
            return Datum(**resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_dataset_status(
        self,
        name: str,
    ) -> Union[TableStatus, None]:
        """
        Get the state of a given dataset.

        Parameters
        ----------
        name : str
            The name of the dataset we want to fetch the state of.

        Returns
        ------
        TableStatus | None
            The state of the dataset, or 'None' if the dataset does not exist.
        """
        try:
            return self.conn.get_dataset_status(name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_dataset_summary(self, name: str) -> DatasetSummary:
        """
        Gets the summary of a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to create a summary for.

        Returns
        -------
        DatasetSummary
            A dataclass containing the dataset summary.
        """
        return DatasetSummary(**self.conn.get_dataset_summary(name))

    def delete_dataset(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the dataset was deleted.
        """
        self.conn.delete_dataset(name)
        if timeout:
            for _ in range(timeout):
                if self.get_dataset(name) is None:
                    break
                else:
                    time.sleep(1)
            else:
                raise TimeoutError(
                    "Dataset wasn't deleted within timeout interval"
                )

    def create_model(
        self,
        model: Union[Model, dict],
    ):
        """
        Creates a model.

        Parameters
        ----------
        model : valor.Model
            The model to create.
        """
        if isinstance(model, Model):
            model = encode_api_format(model)
        self.conn.create_model(model)

    def create_predictions(
        self,
        dataset: Dataset,
        model: Model,
        predictions: List[Prediction],
    ) -> None:
        """
        Creates predictions.

        Parameters
        ----------
        dataset : valor.Dataset
            The dataset that is being operated over.
        model : valor.Model
            The model making the prediction.
        predictions : List[valor.Prediction]
            The predictions to create.
        """
        predictions_json = []
        for prediction in predictions:
            if not isinstance(prediction, Prediction):
                raise TypeError(
                    f"Expected prediction to be of type 'valor.Prediction' not '{type(prediction)}'."
                )
            if not isinstance(prediction.annotations._value, list):
                raise TypeError
            prediction_dict = encode_api_format(prediction)
            prediction_dict["datum"]["dataset_name"] = dataset.get_name()
            prediction_dict["model_name"] = model.get_name()
            predictions_json.append(prediction_dict)
        self.conn.create_predictions(predictions_json)

    def get_prediction(
        self,
        dataset: Union[Dataset, str],
        model: Union[Model, str],
        datum: Union[Datum, str],
    ) -> Union[Prediction, None]:
        """
        Get a particular prediction.

        Parameters
        ----------
        dataset: Union[Dataset, str]
            The dataset the datum belongs to.
        model: Union[Model, str]
            The model that made the prediction.
        datum: Union[Datum, str]
            The desired datum.

        Returns
        ----------
        Union[Prediction, None]
            The matching prediction or 'None' if it doesn't exist.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.get_name() if isinstance(model, Model) else model
        datum_uid = datum.get_uid() if isinstance(datum, Datum) else datum
        try:
            resp = self.conn.get_prediction(
                dataset_name=dataset_name,
                model_name=model_name,
                datum_uid=datum_uid,
            )
            resp = decode_api_format(resp)
            return Prediction(**resp)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def finalize_inferences(
        self, dataset: Union[Dataset, str], model: Union[Model, str]
    ) -> None:
        """
        Finalizes a model-dataset pairing such that new predictions cannot be added to it.
        """
        dataset_name = (
            dataset.get_name() if isinstance(dataset, Dataset) else dataset
        )
        model_name = model.get_name() if isinstance(model, Model) else model
        return self.conn.finalize_inferences(
            dataset_name=dataset_name,
            model_name=model_name,
        )

    def get_model(
        self,
        name: str,
    ) -> Union[Model, None]:
        """
        Gets a model by name.

        Parameters
        ----------
        name : str
            The name of the model to fetch.

        Returns
        -------
        Union[valor.Model, None]
            A Model with matching name or 'None' if one doesn't exist.
        """
        try:
            resp = decode_api_format(self.conn.get_model(name))
            model = Model(**resp)
            model.add_connection(self.conn)
            return model
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_models(
        self,
        filter_by: Optional[FilterType] = None,
    ) -> List[Model]:
        """
        Get all models using an optional filter.

        Parameters
        ----------
        filter_by : FilterType, optional
            Optional constraints to filter by.

        Returns
        ------
        List[valor.Model]
            A list of models.
        """
        filter_ = _format_filter(filter_by)
        if isinstance(filter_, Filter):
            filter_ = asdict(filter_)
        model_list = []
        for kwargs in self.conn.get_models(filter_):
            kwargs = decode_api_format(kwargs)
            model = Model(**kwargs)
            model.add_connection(self.conn)
            model_list.append(model)
        return model_list

    def get_model_status(
        self,
        dataset_name: str,
        model_name: str,
    ) -> Optional[TableStatus]:
        """
        Get the state of a given model over a dataset.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset that the model is operating over.
        model_name : str
            The name of the model we want to fetch the state of.

        Returns
        ------
        Union[TableStatus, None]
            The state of the model or 'None' if the model doesn't exist.
        """
        try:
            return self.conn.get_model_status(dataset_name, model_name)
        except ClientException as e:
            if e.status_code == 404:
                return None
            raise e

    def get_model_eval_requests(
        self, model: Union[Model, str]
    ) -> List[Evaluation]:
        """
        Get all evaluations that have been created for a model.

        This does not return evaluation results.

        `GET` endpoint.

        Parameters
        ----------
        model : str
            The model to search by.

        Returns
        -------
        List[Evaluation]
            A list of evaluations.
        """
        model_name = model.get_name() if isinstance(model, Model) else model
        return [
            Evaluation(**evaluation, connection=self.conn)
            for evaluation in self.conn.get_model_eval_requests(model_name)
        ]

    def delete_model(self, name: str, timeout: int = 0) -> None:
        """
        Deletes a model.

        Parameters
        ----------
        name : str
            The name of the model to be deleted.
        timeout : int
            The number of seconds to wait in order to confirm that the model was deleted.
        """
        self.conn.delete_model(name)
        if timeout:
            for _ in range(timeout):
                if self.get_model(name) is None:
                    break
                else:
                    time.sleep(1)
            else:
                raise TimeoutError(
                    "Model wasn't deleted within timeout interval"
                )

    def get_evaluations(
        self,
        *,
        evaluation_ids: Optional[List[int]] = None,
        models: Union[List[Model], List[str], None] = None,
        datasets: Union[List[Dataset], List[str], None] = None,
    ) -> List[Evaluation]:
        """
        Returns all evaluations associated with user-supplied dataset and/or model names.

        Parameters
        ----------
        evaluation_ids : List[int], optional.
            A list of job IDs to return metrics for.
        models : Union[List[valor.Model], List[str]], optional
            A list of model names that we want to return metrics for.
        datasets : Union[List[valor.Dataset], List[str]], optional
            A list of dataset names that we want to return metrics for.

        Returns
        -------
        List[valor.Evaluation]
            A list of evaluations.
        """
        if isinstance(datasets, list):
            datasets = [
                element.get_name() if isinstance(element, Dataset) else element
                for element in datasets
            ]
        if isinstance(models, list):
            models = [
                element.get_name() if isinstance(element, Model) else element
                for element in models
            ]
        return [
            Evaluation(connection=self.conn, **evaluation)
            for evaluation in self.conn.get_evaluations(
                evaluation_ids=evaluation_ids,
                models=models,
                datasets=datasets,
            )
        ]

    def evaluate(self, request: EvaluationRequest) -> List[Evaluation]:
        """
        Creates as many evaluations as necessary to fulfill the request.

        Parameters
        ----------
        request : schemas.EvaluationRequest
            The requested evaluation parameters.

        Returns
        -------
        List[Evaluation]
            A list of evaluations that meet the parameters.
        """
        return [
            Evaluation(**evaluation)
            for evaluation in self.conn.evaluate(request)
        ]



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from valor.enums import TaskType
from valor.schemas.symbolic.operators import (
    And,
    AppendableFunction,
    Function,
    Inside,
    Intersects,
    IsNotNull,
    IsNull,
    Negate,
    OneArgumentFunction,
    Or,
    Outside,
    TwoArgumentFunction,
    Xor,
)
from valor.schemas.symbolic.types import (
    Date,
    DateTime,
    Duration,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    Time,
    Variable,
)


@dataclass
class Constraint:
    """
    Represents a constraint with a value and an operator.

    Attributes:
        value : Any
            The value associated with the constraint.
        operator : str
            The operator used to define the constraint.
    """

    value: Any
    operator: str


def _convert_symbol_to_attribute_name(symbol_name):
    map_sym_to_attr = {
        "dataset.name": "dataset_names",
        "dataset.metadata": "dataset_metadata",
        "model.name": "model_names",
        "model.metadata": "model_metadata",
        "datum.uid": "datum_uids",
        "datum.metadata": "datum_metadata",
        "annotation.task_type": "task_types",
        "annotation.metadata": "annotation_metadata",
        "annotation.bounding_box": "require_bounding_box",
        "annotation.bounding_box.area": "bounding_box_area",
        "annotation.polygon": "require_polygon",
        "annotation.polygon.area": "polygon_area",
        "annotation.raster": "require_raster",
        "annotation.raster.area": "raster_area",
        "annotation.labels": "labels",
        "label.id": "label_ids",
        "label.key": "label_keys",
        "label.score": "label_scores",
    }
    return map_sym_to_attr[symbol_name]


def _convert_expression_to_constraint(expr: Function):
    # extract value
    if isinstance(expr, TwoArgumentFunction):
        variable = expr.rhs
        if isinstance(
            variable,
            (
                Point,
                MultiPoint,
                LineString,
                MultiLineString,
                Polygon,
                MultiPolygon,
            ),
        ):
            value = {
                "type": type(variable).__name__.capitalize(),
                "coordinates": variable.get_value(),
            }
        elif isinstance(variable, (DateTime, Date, Time, Duration)):
            value = {type(variable).__name__.lower(): variable.encode_value()}
        else:
            value = variable.encode_value()
    else:
        value = None

    # extract operator
    if hasattr(expr, "_operator") and expr._operator is not None:
        op = expr._operator
    elif isinstance(expr, Inside):
        op = "inside"
    elif isinstance(expr, Intersects):
        op = "intersect"
    elif isinstance(expr, Outside):
        op = "outside"
    elif isinstance(expr, IsNotNull):
        op = "exists"
    elif isinstance(expr, IsNull):
        op = "is_none"
    else:
        raise NotImplementedError(
            f"Function '{type(expr)}' has not been implemented by the API."
        )

    return Constraint(value=value, operator=op)


def _scan_one_arg_function(fn: OneArgumentFunction):
    if not fn.arg.is_symbolic:
        raise ValueError(
            "Single argument functions should take a symbol as input."
        )


def _scan_two_arg_function(fn: TwoArgumentFunction):
    if not isinstance(fn.lhs, Variable) or not isinstance(fn.rhs, Variable):
        raise ValueError("Nested arguments are currently unsupported.")
    elif not fn.lhs.is_symbolic:
        raise ValueError(
            f"Values on the lhs of an operator are currently unsupported. {fn.lhs}"
        )
    elif not fn.rhs.is_value:
        raise ValueError(
            f"Symbols on the rhs of an operator are currently unsupported. {fn.rhs}"
        )


def _scan_appendable_function(fn: AppendableFunction):
    if not isinstance(fn, (And, Or)):
        raise ValueError(
            f"Operation '{type(fn)}' is currently unsupported by the API."
        )

    symbols = set()
    for arg in fn._args:
        if not isinstance(fn, Function):
            raise ValueError(
                f"Expected a function but received value with type '{type(fn)}'"
            )

        # scan for nested logic
        if isinstance(arg, (Or, And, Xor, Negate)):
            raise NotImplementedError

        # scan for symbol/value positioning
        if isinstance(arg, OneArgumentFunction):
            _scan_one_arg_function(arg)
        elif isinstance(arg, TwoArgumentFunction):
            _scan_two_arg_function(arg)

        symbols.add(arg._args[0].get_symbol())

    # check that only one symbol was defined per statement
    if len(symbols) > 1:
        raise ValueError(
            f"Defining more than one variable per statement is currently unsupported. {symbols}"
        )
    symbol = list(symbols)[0]

    # check that symbol is compatible with the logical operation
    if isinstance(fn, And) and not (
        symbol._name in Filter._supports_and()
        or symbol._attribute in Filter._supports_and()
    ):
        raise ValueError(
            f"Symbol '{str(symbol)}' currently does not support the 'AND' operation."
        )
    elif isinstance(fn, Or) and not (
        symbol._name in Filter._supports_or()
        or symbol._attribute in Filter._supports_or()
    ):
        raise ValueError(
            f"Symbol '{str(symbol)}' currently does not support the 'AND' operation."
        )


def _parse_listed_expressions(flist):
    expressions = {}
    for row in flist:
        if not isinstance(row, Function):
            raise ValueError(
                f"Expected a function but received value with type '{type(row)}'"
            )
        elif isinstance(row, AppendableFunction):
            _scan_appendable_function(row)
            symbol = row._args[0]._args[0].get_symbol()
            constraints = [
                _convert_expression_to_constraint(arg) for arg in row._args
            ]
        elif isinstance(row, TwoArgumentFunction):
            _scan_two_arg_function(row)
            symbol = row.lhs.get_symbol()
            constraints = [_convert_expression_to_constraint(row)]
        elif isinstance(row, OneArgumentFunction):
            _scan_one_arg_function(row)
            symbol = row.arg.get_symbol()
            constraints = [_convert_expression_to_constraint(row)]
        else:
            raise NotImplementedError

        symbol_name = f"{symbol._owner}.{symbol._name}"
        if symbol._attribute:
            symbol_name += f".{symbol._attribute}"
        attribute_name = _convert_symbol_to_attribute_name(symbol_name)
        if symbol._key:
            if attribute_name not in expressions:
                expressions[attribute_name] = dict()
            if symbol._key not in expressions[attribute_name]:
                expressions[attribute_name][symbol._key] = list()
            expressions[attribute_name][symbol._key] += constraints
        else:
            if attribute_name not in expressions:
                expressions[attribute_name] = list()
            expressions[attribute_name] += constraints

    return expressions


@dataclass
class Filter:
    """
    Used to filter Evaluations according to specific, user-defined criteria.

    Attributes
    ----------
    dataset_names : List[str], optional
        A list of `Dataset` names to filter on.
    dataset_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Dataset` metadata to filter on.
    model_names : List[str], optional
        A list of `Model` names to filter on.
    model_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Model` metadata to filter on.
    datum_uids : List[str], optional
        A list of `Datum` UIDs to filter on.
    datum_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Datum` metadata to filter on.
    task_types : List[TaskType], optional
        A list of task types to filter on.
    annotation_metadata : Dict[str, List[Constraint]], optional
        A dictionary of `Annotation` metadata to filter on.
    require_bounding_box : bool, optional
        A toggle for filtering by bounding boxes.
    bounding_box_area : bool, optional
        An optional constraint to filter by bounding box area.
    require_polygon : bool, optional
        A toggle for filtering by polygons.
    polygon_area : bool, optional
        An optional constraint to filter by polygon area.
    require_raster : bool, optional
        A toggle for filtering by rasters.
    raster_area : bool, optional
        An optional constraint to filter by raster area.
    labels : List[Label], optional
        A list of `Labels' to filter on.
    label_ids : List[int], optional
        A list of label row id's.
    label_keys : List[str], optional
        A list of `Label` keys to filter on.
    label_scores : List[Constraint], optional
        A list of `Constraints` which are used to filter `Evaluations` according to the `Model`'s prediction scores.

    Raises
    ------
    TypeError
        If `value` isn't of the correct type.
    ValueError
        If the `operator` doesn't match one of the allowed patterns.
    """

    # datasets
    dataset_names: Optional[List[str]] = None
    dataset_metadata: Optional[Dict[str, List[Constraint]]] = None

    # models
    model_names: Optional[List[str]] = None
    model_metadata: Optional[Dict[str, List[Constraint]]] = None

    # datums
    datum_uids: Optional[List[str]] = None
    datum_metadata: Optional[Dict[str, List[Constraint]]] = None

    # annotations
    task_types: Optional[List[TaskType]] = None
    annotation_metadata: Optional[Dict[str, List[Constraint]]] = None

    # geometries
    require_bounding_box: Optional[bool] = None
    bounding_box_area: Optional[List[Constraint]] = None
    require_polygon: Optional[bool] = None
    polygon_area: Optional[List[Constraint]] = None
    require_raster: Optional[bool] = None
    raster_area: Optional[List[Constraint]] = None

    # labels
    labels: Optional[List[Dict[str, str]]] = None
    label_ids: Optional[List[int]] = None
    label_keys: Optional[List[str]] = None
    label_scores: Optional[List[Constraint]] = None

    @staticmethod
    def _supports_and():
        return {
            "area",
            "score",
            "metadata",
        }

    @staticmethod
    def _supports_or():
        return {
            "name",
            "uid",
            "task_type",
            "labels",
            "keys",
        }

    def __post_init__(self):
        def _unpack_metadata(metadata: Optional[dict]) -> Union[dict, None]:
            if metadata is None:
                return None
            for k, vlist in metadata.items():
                metadata[k] = [
                    v if isinstance(v, Constraint) else Constraint(**v)
                    for v in vlist
                ]
            return metadata

        # unpack metadata
        self.dataset_metadata = _unpack_metadata(self.dataset_metadata)
        self.model_metadata = _unpack_metadata(self.model_metadata)
        self.datum_metadata = _unpack_metadata(self.datum_metadata)
        self.annotation_metadata = _unpack_metadata(self.annotation_metadata)

        def _unpack_list(
            vlist: Optional[list], object_type: type
        ) -> Optional[list]:
            def _handle_conversion(v, object_type):
                if object_type is Constraint:
                    return object_type(**v)
                else:
                    return object_type(v)

            if vlist is None:
                return None

            return [
                (
                    v
                    if isinstance(v, object_type)
                    else _handle_conversion(v=v, object_type=object_type)
                )
                for v in vlist
            ]

        # unpack tasktypes
        self.task_types = _unpack_list(self.task_types, TaskType)

        # unpack area
        self.bounding_box_area = _unpack_list(
            self.bounding_box_area, Constraint
        )
        self.polygon_area = _unpack_list(self.polygon_area, Constraint)
        self.raster_area = _unpack_list(self.raster_area, Constraint)

        # scores
        self.label_scores = _unpack_list(self.label_scores, Constraint)

    @classmethod
    def create(cls, expressions: List[Any]):
        """
        Parses a list of `BinaryExpression` to create a `schemas.Filter` object.

        Parameters
        ----------
        expressions: Sequence[Union[BinaryExpression, Sequence[BinaryExpression]]]
            A list of (lists of) `BinaryExpressions' to parse into a `Filter` object.
        """

        constraints = _parse_listed_expressions(expressions)

        # create filter
        filter_request = cls()

        # metadata constraints
        for attr in [
            "dataset_metadata",
            "model_metadata",
            "datum_metadata",
            "annotation_metadata",
            "bounding_box_area",
            "polygon_area",
            "raster_area",
            "label_scores",
        ]:
            if attr in constraints:
                setattr(filter_request, attr, constraints[attr])

        # boolean constraints
        for attr in [
            "require_bounding_box",
            "require_polygon",
            "require_raster",
        ]:
            if attr in constraints:
                for constraint in constraints[attr]:
                    if constraint.operator == "exists":
                        setattr(filter_request, attr, True)
                    elif constraint.operator == "is_none":
                        setattr(filter_request, attr, False)

        # equality constraints
        for attr in [
            "dataset_names",
            "model_names",
            "datum_uids",
            "task_types",
            "label_keys",
        ]:
            if attr in constraints:
                setattr(
                    filter_request,
                    attr,
                    [expr.value for expr in constraints[attr]],
                )

        # edge case - label list
        if "labels" in constraints:
            setattr(
                filter_request,
                "labels",
                [
                    {label["key"]: label["value"]}
                    for labels in constraints["labels"]
                    for label in labels.value
                ],
            )

        return filter_request



import numpy as np
import PIL.Image
import pytest

from valor import Annotation, GroundTruth, Label
from valor.enums import TaskType
from valor.metatypes import ImageMetadata
from valor.schemas import MultiPolygon, Polygon, Raster
from valor.viz import (
    _polygons_to_binary_mask,
    create_combined_segmentation_mask,
    draw_detections_on_image,
)


@pytest.fixture
def bounding_poly() -> Polygon:
    return Polygon(
        [
            [
                (100, 100),
                (200, 100),
                (200, 200),
                (100, 200),
                (100, 100),
            ]
        ]
    )


@pytest.fixture
def poly1(bounding_poly: Polygon) -> Polygon:
    return Polygon(
        [
            bounding_poly.get_value()[0],
            [
                (150, 120),
                (180, 120),
                (180, 140),
                (150, 140),
                (150, 120),
            ],
        ]
    )


def test__polygons_to_binary_mask(poly1):
    poly2 = Polygon(
        [
            [
                (10, 15),
                (20, 15),
                (20, 20),
                (10, 20),
                (10, 15),
            ]
        ]
    )

    mask = _polygons_to_binary_mask([poly1, poly2], 500, 600)

    area_poly1 = (200 - 100 + 1) * (200 - 100 + 1) - (180 - 150 + 1) * (
        140 - 120 + 1
    )

    area_poly2 = (20 - 10 + 1) * (20 - 15 + 1)

    assert mask.sum() == area_poly1 + area_poly2


def test_create_combined_segmentation_mask(poly1: Polygon):
    image = ImageMetadata.create(uid="uid", height=200, width=200).datum

    gt1 = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                task_type=TaskType.OBJECT_DETECTION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v2"),
                    Label(key="k3", value="v3"),
                ],
                raster=Raster.from_geometry(
                    MultiPolygon([poly1.get_value()]),
                    height=2,
                    width=2,
                ),
            ),
            Annotation(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v3"),
                ],
                raster=Raster.from_numpy(
                    np.array([[True, False], [False, True]]),
                ),
            ),
        ],
    )

    gt_with_size_mismatch = GroundTruth(
        datum=image,
        annotations=[
            Annotation(
                task_type=TaskType.OBJECT_DETECTION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v2"),
                    Label(key="k3", value="v3"),
                ],
                raster=Raster.from_geometry(
                    MultiPolygon([poly1.get_value()]),
                    height=20,
                    width=20,
                ),
            ),
            Annotation(
                task_type=TaskType.SEMANTIC_SEGMENTATION,
                labels=[
                    Label(key="k1", value="v1"),
                    Label(key="k2", value="v3"),
                ],
                raster=Raster.from_numpy(
                    np.array([[True, False], [False, True]]),
                ),
            ),
        ],
    )

    # test that a size mistmatch between rasters is caught
    with pytest.raises(ValueError) as exc_info:
        create_combined_segmentation_mask(
            gt_with_size_mismatch,
            label_key="k2",
            task_type=TaskType.SEMANTIC_SEGMENTATION,
        )
    assert "(20, 20) != (2, 2)" in str(exc_info)

    # check get an error since "k3" isn't a label key in seg2
    with pytest.raises(RuntimeError) as exc_info:
        create_combined_segmentation_mask(
            gt1,
            label_key="k3",
            task_type=TaskType.SEMANTIC_SEGMENTATION,
        )
    assert "doesn't have a label" in str(exc_info)

    # should have one distinct (non-black) color
    combined_mask, legend = create_combined_segmentation_mask(
        gt1,
        label_key="k1",
        task_type=TaskType.SEMANTIC_SEGMENTATION,
    )
    combined_mask = np.array(combined_mask)
    # check that we get two unique RGB values (black and one color for label value "v1")
    assert combined_mask.shape == (2, 2, 3)
    assert len(legend) == 1  # background color 'black' is not counted

    # should have two distinct (non-black) color
    combined_mask, legend = create_combined_segmentation_mask(
        gt1, label_key="k2"
    )
    combined_mask = np.array(combined_mask)
    # check that we get three unique RGB values (black and two colors for label values "v2" and "v3")
    assert combined_mask.shape == (2, 2, 3)
    assert len(legend) == 2  # background color 'black' is not counted


def test_draw_detections_on_image(bounding_poly: Polygon):
    detections = [
        GroundTruth(
            datum=ImageMetadata.create("test", 300, 300).datum,
            annotations=[
                Annotation(
                    task_type=TaskType.OBJECT_DETECTION,
                    labels=[Label(key="k", value="v")],
                    polygon=Polygon(bounding_poly.get_value()),
                )
            ],
        ),
    ]
    img = PIL.Image.new("RGB", (300, 300))

    img = draw_detections_on_image(detections, img)

    assert img.size == (300, 300)

    # check unique colors only have red component
    unique_rgb = np.unique(np.array(img).reshape(-1, 3), axis=0)
    assert unique_rgb[:, 1:].sum() == 0



from unittest.mock import patch

import pytest

from valor.client import _chunk_list, connect, get_connection, reset_connection
from valor.exceptions import (
    ClientAlreadyConnectedError,
    ClientConnectionFailed,
    ClientNotConnectedError,
)


def test__chunk_list():
    # edge case
    data = [{"key": "value"}]
    results = _chunk_list(json_list=data, chunk_size_bytes=10)

    assert results == [data]

    # 100 elements with an average element size of 23.8
    data = [{f"key_{i}": f"value_{i}"} for i in range(100)]

    # standard case
    results = _chunk_list(json_list=data, chunk_size_bytes=1000)
    assert (
        len(results) == 4
    )  # recursively chunked once, which added an extra split
    assert [len(x) for x in results] == [42, 41, 1, 16]

    # edge case with small chunk size
    results = _chunk_list(json_list=data, chunk_size_bytes=1)
    assert results == [[x] for x in data]


@patch("valor.client.ClientConnection")
def test_connect(ClientConnection):
    connect(host="host")
    ClientConnection.assert_called_once()

    with pytest.raises(ClientAlreadyConnectedError):
        connect(host="host")

    connect(host="host", reconnect=True)

    ClientConnection.side_effect = ClientConnectionFailed("testing")
    with pytest.raises(ClientConnectionFailed) as e:
        connect(host="host", reconnect=True)
    assert "testing" in str(e)


@patch("valor.client.ClientConnection")
def test_get_connection(ClientConnection):
    reset_connection()

    with pytest.raises(ClientNotConnectedError):
        get_connection()

    connect(host="host")
    ClientConnection.assert_called_once()


@patch("valor.client.ClientConnection")
def test_reset_connection(ClientConnection):
    connect(host="host", reconnect=True)
    assert get_connection() is not None

    reset_connection()

    with pytest.raises(ClientNotConnectedError):
        get_connection()
    connect(host="host")  # test without reconnect arg
    assert get_connection() is not None



from .evaluation import EvaluationParameters, EvaluationRequest
from .filters import Constraint, Filter
from .symbolic.collections import Annotation, Datum, Label, StaticCollection
from .symbolic.operators import (
    And,
    Eq,
    Ge,
    Gt,
    Inside,
    Intersects,
    IsNotNull,
    IsNull,
    Le,
    Lt,
    Ne,
    Negate,
    Or,
    Outside,
    Xor,
)
from .symbolic.types import (
    Bool,
    Box,
    Date,
    DateTime,
    Dictionary,
    Duration,
    Embedding,
    Equatable,
    Float,
    Integer,
    LineString,
    List,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    Quantifiable,
    Raster,
    Spatial,
    String,
    Symbol,
    TaskTypeEnum,
    Time,
    Variable,
)

__all__ = [
    "EvaluationRequest",
    "EvaluationParameters",
    "Filter",
    "Constraint",
    "And",
    "Eq",
    "Ge",
    "Gt",
    "Inside",
    "Intersects",
    "IsNotNull",
    "IsNull",
    "Le",
    "Lt",
    "Ne",
    "Negate",
    "Or",
    "Outside",
    "Xor",
    "Symbol",
    "Variable",
    "Equatable",
    "Quantifiable",
    "Spatial",
    "Bool",
    "Box",
    "Integer",
    "Float",
    "String",
    "DateTime",
    "Date",
    "Time",
    "Duration",
    "StaticCollection",
    "Point",
    "MultiPoint",
    "LineString",
    "MultiLineString",
    "Polygon",
    "MultiPolygon",
    "Raster",
    "TaskTypeEnum",
    "Embedding",
    "List",
    "Dictionary",
    "Label",
    "Annotation",
    "Datum",
]



from typing import Any, Optional, Union

from valor.schemas.symbolic.types import (
    Box,
    Date,
    DateTime,
    Dictionary,
    Duration,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    Raster,
    Time,
    get_type_by_name,
)


def _encode_api_metadata_values(variable):
    if isinstance(
        variable,
        (
            Point,
            MultiPoint,
            LineString,
            MultiLineString,
            Polygon,
            MultiPolygon,
        ),
    ):
        return {
            "geojson": {
                "type": type(variable).__name__.lower(),
                "coordinates": variable.get_value(),
            }
        }
    elif isinstance(variable, (DateTime, Date, Time, Duration)):
        return {type(variable).__name__.lower(): variable.encode_value()}
    else:
        return variable.encode_value()


def _encode_api_metadata(metadata: Dictionary) -> dict:
    return {k: _encode_api_metadata_values(v) for k, v in metadata.items()}


def _encode_api_geometry(geometry: Union[Box, Polygon, MultiPolygon, Raster]):
    if geometry.get_value() is None:
        return None
    elif isinstance(geometry, Box):
        return {
            "polygon": {
                "points": [
                    {"x": pt[0], "y": pt[1]} for pt in geometry.boundary[:-1]
                ]
            }
        }
    elif isinstance(geometry, Polygon):
        return {
            "boundary": {
                "points": [
                    {"x": pt[0], "y": pt[1]} for pt in geometry.boundary[:-1]
                ]
            },
            "holes": [
                {"points": [{"x": pt[0], "y": pt[1]} for pt in hole[:-1]]}
                for hole in geometry.holes
            ],
        }
    elif isinstance(geometry, MultiPolygon):
        return {
            "polygons": [
                _encode_api_geometry(poly) for poly in geometry.to_polygons()
            ]
        }
    elif isinstance(geometry, Raster):
        value = geometry.encode_value()
        if value["geometry"] is not None:
            if Polygon.supports(value["geometry"]):
                value["geometry"] = _encode_api_geometry(
                    Polygon(value["geometry"])
                )
            elif MultiPolygon.supports(value["geometry"]):
                value["geometry"] = _encode_api_geometry(
                    MultiPolygon(value["geometry"])
                )
            else:
                raise ValueError
        return value


def encode_api_format(obj: Any) -> dict:
    """Encodes form client format into api format."""
    json = obj.encode_value()

    # static collection
    if "datum" in json:
        json["datum"] = encode_api_format(obj.datum)
    if "annotations" in json:
        json["annotations"] = [
            encode_api_format(annotation) for annotation in obj.annotations
        ]

    # dictionary
    if "metadata" in json:
        json["metadata"] = _encode_api_metadata(obj.metadata)

    # geometry
    if "bounding_box" in json:
        json["bounding_box"] = _encode_api_geometry(obj.bounding_box)
    if "polygon" in json:
        json["polygon"] = _encode_api_geometry(obj.polygon)
    if "raster" in json:
        json["raster"] = _encode_api_geometry(obj.raster)

    return json


def _decode_api_metadata_values(value: Any):
    if not isinstance(value, dict):
        return value
    elif set(value.keys()) == {"geojson"}:
        obj = get_type_by_name(value["geojson"]["type"])
        return obj.decode_value(value["geojson"]["coordinates"])
    elif len(value) == 1:
        k, v = list(value.items())[0]
        obj = get_type_by_name(k)
        return obj.decode_value(v)
    else:
        raise NotImplementedError(str(value))


def _decode_api_metadata(metadata) -> dict:
    return {k: _decode_api_metadata_values(v) for k, v in metadata.items()}


def _decode_api_geometry(value: Optional[dict]):
    if value is None:
        return None
    elif set(value.keys()) == {"polygon"}:
        return [
            [
                (pt["x"], pt["y"])
                for pt in [
                    *value["polygon"]["points"],
                    value["polygon"]["points"][0],
                ]
            ]
        ]
    elif set(value.keys()) == {"boundary", "holes"}:
        boundary = [
            (pt["x"], pt["y"])
            for pt in [
                *value["boundary"]["points"],
                value["boundary"]["points"][0],
            ]
        ]
        holes = (
            [
                [
                    (pt["x"], pt["y"])
                    for pt in [*hole["points"], hole["points"][0]]
                ]
                for hole in value["holes"]
            ]
            if value["holes"]
            else None
        )
        return [boundary, *holes] if holes else [boundary]
    elif set(value.keys()) == {"mask", "geometry"}:
        if value["geometry"] is not None:
            value["geometry"] = _decode_api_geometry(value["geometry"])
        return Raster.decode_value(value).get_value()  # type: ignore - none case already handled


def decode_api_format(json: dict):
    """Decoded api format into client format."""
    # objects
    if "datum" in json:
        json["datum"] = decode_api_format(json["datum"])
    if "annotations" in json:
        json["annotations"] = [
            decode_api_format(annotation) for annotation in json["annotations"]
        ]
    # data
    if "metadata" in json:
        json["metadata"] = _decode_api_metadata(json["metadata"])
    if "model_name" in json:
        json.pop("model_name")
    if "dataset_name" in json:
        json.pop("dataset_name")
    if "bounding_box" in json:
        json["bounding_box"] = _decode_api_geometry(json["bounding_box"])
    if "polygon" in json:
        json["polygon"] = _decode_api_geometry(json["polygon"])
    if "raster" in json:
        json["raster"] = _decode_api_geometry(json["raster"])

    return json



import typing

import numpy as np
import pytest

from valor.schemas import Box, Embedding, Float, Raster, TaskTypeEnum
from valor.schemas.symbolic.operators import (
    AppendableFunction,
    TwoArgumentFunction,
)


def get_function_name(fn: str) -> str:
    fns = {
        "__eq__": "eq",
        "__ne__": "ne",
        "__and__": "and",
        "__or__": "or",
        "__xor__": "xor",
        "__gt__": "gt",
        "__ge__": "ge",
        "__lt__": "lt",
        "__le__": "le",
        "is_none": "isnull",
        "is_not_none": "isnotnull",
        "intersects": "intersects",
        "inside": "inside",
        "outside": "outside",
    }
    return fns[fn]


def _test_encoding(objcls, value, encoded_value):
    assert (
        objcls(value).to_dict() == objcls.decode_value(encoded_value).to_dict()
    )
    assert encoded_value == objcls(value).encode_value()


def _test_to_dict(objcls, value, type_name: typing.Optional[str] = None):
    type_name = type_name if type_name else objcls.__name__.lower()
    # test __init__
    assert objcls(value).to_dict() == {
        "type": type_name,
        "value": objcls(value).encode_value(),
    }
    # test value
    assert objcls(value).to_dict() == {
        "type": type_name,
        "value": objcls(value).encode_value(),
    }
    # test symbolic
    assert objcls.symbolic().to_dict() == {
        "type": "symbol",
        "value": {
            "owner": None,
            "name": type_name,
            "key": None,
            "attribute": None,
        },
    }


def _test_generic(
    objcls, permutations, op, type_name: typing.Optional[str] = None
):
    """Tests expressions that can only be resolved to JSON."""
    for a, _ in permutations:
        A = objcls.nullable(a)
        C = objcls.symbolic()
        # test variable -> builtin against variable -> variable
        assert (
            C.__getattribute__(op)(a).to_dict()
            == C.__getattribute__(op)(A).to_dict()
        )
        # test commutative propery (this will fail)
        with pytest.raises(AssertionError):
            try:
                # function does not exist in left-operand
                a.__getattribute__(op)(C)
                # function exists, but is not commutative
                if type(a.__getattribute__(op)(A)) not in {objcls, type(a)}:
                    raise AssertionError("NotImplementedType")
            except AttributeError as e:
                raise AssertionError(e)
        # test instance dictionary generation
        _test_to_dict(objcls, a, type_name)
        # test functional dictionary generation
        expr = C.__getattribute__(op)(a)
        expr_dict = expr.to_dict()
        if issubclass(type(expr), AppendableFunction):
            assert len(expr_dict) == 2
            assert expr_dict["op"] == get_function_name(op)
            assert expr_dict["args"] == [
                C.to_dict(),
                A.to_dict(),
            ]
        elif issubclass(type(expr), TwoArgumentFunction):
            assert len(expr_dict) == 3
            assert expr_dict["op"] == get_function_name(op)
            assert expr_dict["lhs"] == C.to_dict()
            assert expr_dict["rhs"] == A.to_dict()
        else:
            raise AssertionError


def _test_resolvable(
    objcls, permutations, op, type_name: typing.Optional[str] = None
):
    type_name = type_name if type_name else objcls.__name__.lower()

    # test expressions that can be simplified to a 'Bool'
    for a, b in permutations:
        A = objcls(a)
        B = objcls(b)

        # determine truth
        truth = a.__getattribute__(op)(b)

        # test variable -> builtin against truth
        assert A.__getattribute__(op)(b).get_value() is truth
        # test variable -> variable against truth
        assert A.__getattribute__(op)(B).get_value() is truth
        # test dictionary generation
        dictA = A.to_dict()
        assert A.get_value() == a
        assert len(dictA) == 2
        assert dictA["type"] == type_name
        assert dictA["value"] == A.encode_value()
    # test expressions that cannot be simplified
    _test_generic(objcls, permutations, op, type_name=type_name)


def _test_unsupported(objcls, permutations, op):
    for a, b in permutations:
        with pytest.raises(AttributeError):
            objcls(a).__getattribute__(op)(b)


def test_score():
    objcls = Float

    # test supported methods
    permutations = [
        (0.9, 0.1),
        (0.9, 0.9),
        (0.1, 0.9),
        (0.1, 0.1),
    ]
    unresolvable_permutations = [
        (0.9, None),
        (None, 0.9),
    ]
    for op in ["__eq__", "__ne__", "__gt__", "__ge__", "__lt__", "__le__"]:
        _test_resolvable(objcls, permutations, op, type_name="float")
        with pytest.raises(TypeError):
            _test_generic(
                objcls, unresolvable_permutations, op, type_name="float"
            )

        with pytest.raises((AssertionError, TypeError)):
            _test_resolvable(
                objcls,
                unresolvable_permutations,
                op,
                type_name="float",
            )

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none().get_value() is True  # type: ignore - always a bool
    assert v1.is_not_none().get_value() is False  # type: ignore - always a bool
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert v2.is_none().get_value() is False  # type: ignore - always a bool
    assert v2.is_not_none().get_value() is True  # type: ignore - always a boolnullable(None).is_not_none().get_value() is False  # type: ignore - always returns bool

    # test unsupported methods
    for op in [
        "__and__",
        "__or__",
        "__xor__",
        "intersects",
        "inside",
        "outside",
    ]:
        _test_unsupported(objcls, permutations, op)

    # test encoding
    _test_encoding(objcls, 0.2, 0.2)


def test_tasktypeenum():
    from valor.enums import TaskType

    objcls = TaskTypeEnum

    # test supported methods
    permutations = [
        (TaskType.CLASSIFICATION, TaskType.CLASSIFICATION),
        (TaskType.CLASSIFICATION, TaskType.OBJECT_DETECTION),
        (TaskType.OBJECT_DETECTION, TaskType.OBJECT_DETECTION),
        (TaskType.OBJECT_DETECTION, TaskType.CLASSIFICATION),
    ]
    for op in ["__eq__", "__ne__"]:
        _test_resolvable(objcls, permutations, op)

    # test unsupported methods
    for op in [
        "__gt__",
        "__ge__",
        "__lt__",
        "__le__",
        "__and__",
        "__or__",
        "__xor__",
        "intersects",
        "inside",
        "outside",
    ]:
        _test_unsupported(objcls, permutations, op)

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none().get_value() is True  # type: ignore - always a bool
    assert v1.is_not_none().get_value() is False  # type: ignore - always a bool
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert v2.is_none().get_value() is False  # type: ignore - always a bool
    assert v2.is_not_none().get_value() is True  # type: ignore - always a bool

    # test encoding
    _test_encoding(
        objcls, TaskType.CLASSIFICATION, TaskType.CLASSIFICATION.value
    )
    _test_encoding(
        objcls, TaskType.OBJECT_DETECTION, TaskType.OBJECT_DETECTION.value
    )
    _test_encoding(
        objcls,
        TaskType.SEMANTIC_SEGMENTATION,
        TaskType.SEMANTIC_SEGMENTATION.value,
    )
    _test_encoding(objcls, TaskType.EMBEDDING, TaskType.EMBEDDING.value)


def test_box():
    objcls = Box
    value = [[(0, 2), (1, 2), (1, 3), (0, 3), (0, 2)]]
    other = [[(1, 2), (2, 2), (2, 3), (1, 3), (1, 2)]]

    # test __init__
    assert objcls(value).get_value() == value

    # test 'from_extrema' classmethod
    assert objcls.from_extrema(0, 1, 2, 3).get_value() == value

    # test dictionary generation
    assert objcls.from_extrema(0, 1, 2, 3).to_dict() == {
        "type": "box",
        "value": value,
    }

    # test permutations
    permutations = [
        (value, value),
        (value, other),
        (other, other),
        (other, value),
    ]
    for op in ["intersects", "inside", "outside"]:
        _test_generic(objcls, permutations, op)

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none().get_value() is True  # type: ignore - always a bool
    assert v1.is_not_none().get_value() is False  # type: ignore - always a bool
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert v2.is_none().get_value() is False  # type: ignore - always a bool
    assert v2.is_not_none().get_value() is True  # type: ignore - always a bool

    # test unsupported methods
    for op in [
        "__eq__",
        "__ne__",
        "__gt__",
        "__ge__",
        "__lt__",
        "__le__",
        "__and__",
        "__or__",
        "__xor__",
    ]:
        _test_unsupported(objcls, permutations, op)

    # test encoding
    _test_encoding(objcls, value, value)

    # test validate box must define 5 points with first == last
    with pytest.raises(ValueError):
        Box([[(0, 0)]])
    with pytest.raises(ValueError):
        Box(value[:-1])
    value[0][-1] = (10, 10)
    with pytest.raises(ValueError):
        Box(value)


def test_raster():
    objcls = Raster

    bitmask1 = np.full((10, 10), True)
    bitmask2 = np.full((10, 10), False)
    geom = Box.from_extrema(0, 1, 2, 3)

    value = {"mask": bitmask1, "geometry": None}
    other = {"mask": bitmask2, "geometry": geom.get_value()}

    encoded_value = {
        "mask": "iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKAQAAAAClSfIQAAAAEElEQVR4nGP8f5CJgQEXAgAzSQHUW1CW8QAAAABJRU5ErkJggg==",
        "geometry": None,
    }

    # test encoding
    _test_encoding(objcls, value, encoded_value)

    # test permutations
    permutations = [
        (value, value),
        (value, other),
        (other, other),
        (other, value),
    ]
    for op in ["intersects", "inside", "outside"]:
        _test_generic(objcls, permutations, op)

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none().get_value() is True  # type: ignore - always a bool
    assert v1.is_not_none().get_value() is False  # type: ignore - always a bool
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert v2.is_none().get_value() is False  # type: ignore - always a bool
    assert v2.is_not_none().get_value() is True  # type: ignore - always a bool

    # test 'from_numpy' classmethod
    assert Raster.from_numpy(bitmask1).to_dict() == Raster(value).to_dict()

    # test 'from_geometry' classmethod
    assert (
        Raster.from_geometry(geom, 10, 10).to_dict() == Raster(other).to_dict()
    )

    # test type validation
    with pytest.raises(TypeError):
        Raster(123)  # type: ignore
    with pytest.raises(ValueError):
        Raster({})
    with pytest.raises(TypeError):
        Raster({"mask": 123, "geometry": None})  # type: ignore
    with pytest.raises(ValueError) as e:
        Raster({"mask": np.zeros((10,)), "geometry": None})
    assert "2d arrays" in str(e)
    with pytest.raises(ValueError) as e:
        Raster({"mask": np.zeros((10, 10, 10)), "geometry": None})
    assert "2d arrays" in str(e)
    with pytest.raises(ValueError) as e:
        Raster({"mask": np.zeros((10, 10)), "geometry": None})
    assert "bool" in str(e)
    with pytest.raises(TypeError):
        Raster({"mask": bitmask1, "geometry": 123})  # type: ignore

    # test property 'area'
    assert objcls.symbolic().area.is_symbolic
    assert objcls.symbolic().area.to_dict() == {
        "type": "symbol",
        "value": {
            "owner": None,
            "name": objcls.__name__.lower(),
            "key": None,
            "attribute": "area",
        },
    }

    # test property 'area' is not available to values
    with pytest.raises(ValueError):
        objcls(value).area

    # test property 'array'
    assert (bitmask1 == Raster(value).array).all()
    with pytest.warns(RuntimeWarning):
        Raster(other).array

    # test property 'array' is not available to symbols
    with pytest.raises(TypeError):
        Raster.symbolic().array


def test_embedding():
    objcls = Embedding
    value = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    other = [5, 6, 6, 7, 8, 9, 0, 1, 2, 3]

    # test __init__
    assert objcls(value).get_value() == value

    # test dictionary generation
    assert objcls(value).to_dict() == {
        "type": "embedding",
        "value": value,
    }

    # test permutations
    permutations = [
        (value, value),
        (value, other),
        (other, other),
        (other, value),
    ]
    for op in ["intersects", "inside", "outside"]:
        _test_generic(objcls, permutations, op)

    # test nullable
    v1 = objcls.nullable(None)
    assert v1.get_value() is None
    assert v1.is_none().get_value() is True  # type: ignore - always a bool
    assert v1.is_not_none().get_value() is False  # type: ignore - always a bool
    v2 = objcls.nullable(permutations[0][0])
    assert v2.get_value() is not None
    assert v2.is_none().get_value() is False  # type: ignore - always a bool
    assert v2.is_not_none().get_value() is True  # type: ignore - always a bool

    # test unsupported methods
    for op in [
        "__eq__",
        "__ne__",
        "__gt__",
        "__ge__",
        "__lt__",
        "__le__",
        "__and__",
        "__or__",
        "__xor__",
    ]:
        _test_unsupported(objcls, permutations, op)

    # test encoding
    _test_encoding(objcls, value, value)


def test_label():
    pass


def test_annotation():
    pass


def test_datum():
    pass



from typing import Any

import numpy as np


def is_numeric(value: Any) -> bool:
    """
    Checks whether the value input is a numeric type.

    Parameters
    ----------
    value : Any
        The value to check.

    Returns
    -------
    bool
        Whether the value is a number.
    """
    return isinstance(value, (int, float, np.floating))


def is_float(value: Any) -> bool:
    """
    Checks whether the value input is a floating point type.

    Parameters
    ----------
    value : Any
        The value to check.

    Returns
    -------
    bool
        Whether the value is a floating point number.
    """
    return isinstance(value, (float, np.floating))


def is_geojson(value: Any) -> bool:
    """
    Checks if value is an instance of geojson.

    Parameters
    ----------
    value : Any
        The value to check.

    Returns
    -------
    bool
        Whether the value conforms to the geojson standard.
    """
    if not isinstance(value, dict):
        return False
    elif set(value.keys()) != {"type", "coordinates"}:
        return False
    elif value["type"] not in {"Point", "Polygon", "MultiPolygon"}:
        return False
    elif not isinstance(value["coordinates"], list):
        return False
    else:
        return True



from enum import Enum


class AnnotationType(str, Enum):
    NONE = "none"
    BOX = "box"
    POLYGON = "polygon"
    MULTIPOLYGON = "multipolygon"
    RASTER = "raster"


class TaskType(str, Enum):
    SKIP = "skip"
    EMPTY = "empty"
    CLASSIFICATION = "classification"
    OBJECT_DETECTION = "object-detection"
    SEMANTIC_SEGMENTATION = "semantic-segmentation"
    EMBEDDING = "embedding"


class TableStatus(str, Enum):
    CREATING = "creating"
    FINALIZED = "finalized"
    DELETING = "deleting"


class EvaluationStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    DONE = "done"
    FAILED = "failed"
    DELETING = "deleting"



from typing import Any


class Function:
    """Base class for defining a function."""

    _operator = None

    def __init__(self, *args) -> None:
        for arg in args:
            if not hasattr(arg, "to_dict"):
                raise ValueError(
                    "Functions can only take arguments that have a 'to_dict' method defined."
                )
        self._args = list(args)

    def __repr__(self):
        args = ", ".join([arg.__repr__() for arg in self._args])
        return f"{type(self).__name__}({args})"

    def __str__(self):
        values = [arg.__repr__() for arg in self._args]
        if self._operator is None:
            args = ", ".join(values)
            return f"{type(self).__name__}({args})"
        else:
            args = f" {self._operator} ".join(values)
            return f"({args})"

    def __and__(self, other: Any):
        return And(self, other)

    def __or__(self, other: Any):
        return Or(self, other)

    def __xor__(self, other: Any):
        return Xor(self, other)

    def __invert__(self):
        return Negate(self)

    def to_dict(self):
        """Encode to a JSON-compatible dictionary."""
        return {
            "op": type(self).__name__.lower(),
            "args": [arg.to_dict() for arg in self._args],
        }


class OneArgumentFunction(Function):
    """Base class for defining single argument functions."""

    def __init__(self, arg) -> None:
        super().__init__(arg)

    @property
    def arg(self):
        """Returns the argument."""
        return self._args[0]

    def to_dict(self):
        """Encode to a JSON-compatible dictionary."""
        return {"op": type(self).__name__.lower(), "arg": self.arg.to_dict()}


class TwoArgumentFunction(Function):
    """Base class for defining two argument functions."""

    def __init__(self, lhs: Any, rhs: Any) -> None:
        self._lhs = lhs
        self._rhs = rhs
        super().__init__(lhs, rhs)

    @property
    def lhs(self):
        """Returns the lhs operand."""
        return self._lhs

    @property
    def rhs(self):
        """Returns the rhs operand."""
        return self._rhs

    def to_dict(self):
        """Encode to a JSON-compatible dictionary."""
        return {
            "op": type(self).__name__.lower(),
            "lhs": self.lhs.to_dict(),
            "rhs": self.rhs.to_dict(),
        }


class AppendableFunction(Function):
    """Base class for defining functions with an unlimited number of arguments."""

    _function = None

    def __init__(self, *args) -> None:
        """
        Appendable function.
        """
        if len(args) < 2:
            raise TypeError(
                f"missing {2 - len(args)} required positional argument"
            )
        flat_args = []
        for arg in args:
            if isinstance(arg, type(self)):
                flat_args += arg._args
            else:
                flat_args.append(arg)
        super().__init__(*flat_args)

    def append(self, value: Any):
        """Appends an argument to the function."""
        self._args.append(value)
        return self


class And(AppendableFunction):
    """Implementation of logical AND (&)."""

    _operator = "&"

    def __and__(self, other: Any):
        self.append(other)
        return self


class Or(AppendableFunction):
    """Implementation of logical OR (|)."""

    _operator = "|"

    def __or__(self, other: Any):
        self.append(other)
        return self


class Xor(AppendableFunction):
    """Implementation of logical XOR (^)."""

    _operator = "^"

    def __xor__(self, other: Any):
        self.append(other)
        return self


class Negate(OneArgumentFunction):
    """Implementation of logical negation (~)."""

    _operator = "~"

    def __invert__(self):
        """Inverts negation so return contents."""
        return self.arg


class IsNull(OneArgumentFunction):
    """Implementation of is null value check."""

    pass


class IsNotNull(OneArgumentFunction):
    """Implementation of is not null value check."""

    pass


class Eq(TwoArgumentFunction):
    """Implementation of the equality operator '=='."""

    _operator = "=="


class Ne(TwoArgumentFunction):
    """Implementation of the inequality operator '!='."""

    _operator = "!="


class Gt(TwoArgumentFunction):
    """Implementation of the greater-than operator '>'."""

    _operator = ">"


class Ge(TwoArgumentFunction):
    """Implementation of the greater-than or equal operator '>='."""

    _operator = ">="


class Lt(TwoArgumentFunction):
    """Implementation of the less-than operator '<'."""

    _operator = "<"


class Le(TwoArgumentFunction):
    """Implementation of the less-than or equal operator '<='."""

    _operator = "<="


class Intersects(TwoArgumentFunction):
    """Implementation of the spatial 'intersects' operator."""

    pass


class Inside(TwoArgumentFunction):
    """Implementation of the spatial 'inside' operator."""

    pass


class Outside(TwoArgumentFunction):
    """Implementation of the spatial 'outside' operator."""

    pass
